# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The boolean `StateTimeSeries` and bit field `StateVector`.

Each bit represents a boolean condition.
Such states are typically the comparison of a `TimeSeries` against some
threshold, where sub-threshold is good and sup-threshold is bad,
for example.

Single `StateTimeSeries` can be bundled together to form `StateVector`
arrays, representing a bit mask of states that combine to make a detailed
statement of instrumental operation
"""

from __future__ import annotations

import os
from contextlib import suppress
from functools import wraps
from math import (
    ceil,
    log2,
)
from typing import (
    TYPE_CHECKING,
    overload,
)

import numpy
from astropy import units

from ..detector import Channel
from ..io.registry import UnifiedReadWriteMethod
from ..segments import Segment
from ..time import Time
from ..types import Array2D
from .connect import (
    StateVectorDictGet,
    StateVectorDictRead,
    StateVectorDictWrite,
    StateVectorGet,
    StateVectorRead,
    StateVectorWrite,
)
from .core import (
    TimeSeriesBase,
    TimeSeriesBaseDict,
    TimeSeriesBaseList,
    as_series_dict_class,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
        Sequence,
    )
    from typing import (
        Literal,
        NoReturn,
        Self,
        SupportsIndex,
        TypeAlias,
    )

    import nds2
    from astropy.units import (
        Quantity,
        Unit,
    )
    from numpy.typing import DTypeLike

    from ..plot import Plot
    from ..segments import (
        DataQualityDict,
        DataQualityFlag,
    )
    from ..time import SupportsToGps
    from ..typing import (
        ArrayLike1D,
        UnitLike,
    )

    BitsInput: TypeAlias = dict[int, str | None] | Sequence[str | None]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "Bits",
    "StateTimeSeries",
    "StateTimeSeriesDict",
    "StateVector",
    "StateVectorDict",
    "StateVectorList",
]


# -- utilities ----------------------------------------------------------------

def _bool_segments(
    array: Iterable[bool | int],
    start: float = 0,
    delta: float = 1,
    minlen: int = 1,
) -> Iterator[Segment]:
    """Yield segments of consecutive `True` values in a boolean array.

    Parameters
    ----------
    array : `iterable`
        An iterable of boolean-castable values.

    start : `float`
        The value of the first sample on the indexed axis
        (e.g.the GPS start time of the array).

    delta : `float`
        The step size on the indexed axis (e.g. sample duration).

    minlen : `int`, optional
        The minimum number of consecutive `True` values for a segment.

    Yields
    ------
    segment : `tuple`
        ``(start + i * delta, start + (i + n) * delta)`` for a sequence
        of ``n`` consecutive True values starting at position ``i``.

    Notes
    -----
    This method is adapted from original code written by Kipp Cannon and
    distributed under GPLv3.

    The datatype of the values returned will be the larger of the types
    of ``start`` and ``delta``.

    Examples
    --------
    >>> print(list(_bool_segments([0, 1, 0, 0, 0, 1, 1, 1, 0, 1]))
    [(1, 2), (5, 8), (9, 10)]
    >>> print(list(_bool_segments([0, 1, 0, 0, 0, 1, 1, 1, 0, 1]
    ...                           start=100., delta=0.1))
    [(100.1, 100.2), (100.5, 100.8), (100.9, 101.0)]
    """
    array = iter(array)
    i = 0
    while True:
        try:  # get next value
            val = next(array)
        except StopIteration:  # end of array
            return

        if val:  # start of new segment
            n = 1  # count consecutive True
            try:
                while next(array):  # run until segment will end
                    n += 1
            except StopIteration:  # have reached the end
                return  # stop
            finally:  # yield segment (including at StopIteration)
                if n >= minlen:  # ... if long enough
                    yield Segment(start + i * delta, start + (i + n) * delta)
            i += n
        i += 1


# -- StateTimeSeries -----------------

class StateTimeSeries(TimeSeriesBase):
    """Boolean array representing a good/bad state determination.

    Parameters
    ----------
    value : array-like
        Input data array.

    t0 : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    dt : `float`, `~astropy.units.Quantity`, optional
        Time between successive samples (seconds), can also be given inversely
        via `sample_rate`.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        The rate of samples per second (Hertz), can also be given inversely via `dt`.

    times : `array-like`
        The complete array of GPS times accompanying the data for this series.
        This argument takes precedence over `t0` and `dt` so should be given
        in place of these if relevant, not alongside.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional
        Choose to copy the input data to new memory.

    subok : `bool`, optional
        Allow passing of sub-classes by the array generator.

    Notes
    -----
    Key methods

    .. autosummary::

       ~StateTimeSeries.to_dqflag
    """

    def __new__(
        cls,
        data: ArrayLike1D,
        t0: SupportsToGps | None = None,
        dt: float | Quantity | None = None,
        sample_rate: float | Quantity | None = None,
        times: ArrayLike1D | None = None,
        channel: Channel | str | None = None,
        name: str | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new StateTimeSeries."""
        if kwargs.pop("unit", None) is not None:
            msg = f"{cls.__name__} does not accept keyword argument 'unit'"
            raise TypeError(msg)

        data = numpy.asarray(data)
        if not isinstance(data, cls):
            data = data.astype(bool, copy=False)

        return super().__new__(
            cls,
            data,
            t0=t0,
            dt=dt,
            sample_rate=sample_rate,
            times=times,
            name=name,
            channel=channel,
            **kwargs,
        )

    # -- unit handling (always dimensionless)

    @property  # type: ignore[misc]
    def unit(self) -> Unit:
        """The unit of this `StateTimeSeries`."""
        return units.dimensionless_unscaled

    def override_unit(
        self,
        unit: UnitLike,
        parse_strict: str = "raise",
    ) -> NoReturn:
        """Override the unit of this `StateTimeSeries`. UNSUPPORTED DO NOT USE."""
        msg = f"overriding units is not supported for {type(self).__name__}"
        raise NotImplementedError(msg)

    # -- math handling (always boolean)

    def __array_ufunc__(
        self,
        function: Callable,
        method: str,
        *inputs,
        **kwargs,
    ) -> Self | Quantity:
        """Handle ufuncs on this `StateTimeSeries`."""
        out = super().__array_ufunc__(function, method, *inputs, **kwargs)
        if out.ndim:
            return out.astype(bool)
        return out

    def diff(
        self,
        n: int = 1,
        axis: int = -1,
    ) -> Self:
        """Return the difference between successive samples.

        The difference is defined as the exclusive-or of the
        previous and current samples.

        Parameters
        ----------
        n : `int`, optional
            Number of times to apply the difference operation,
            defaults to `1`, i.e. the first difference.

        axis : `int`, optional
            Axis along which to compute the difference, defaults to the last
            axis, i.e. the time axis.

        Returns
        -------
        new : `StateTimeSeries`
            A new `StateTimeSeries` containing the difference of the
            current series, with the same metadata as the original.
        """
        slice1 = (slice(1, None),)
        slice2 = (slice(None, -1),)
        new = (self.value[slice1] ^ self.value[slice2]).view(type(self))
        new.__metadata_finalize__(self)
        try:  # shift x0 to the right by one place
            new.x0 = self._xindex[1]
        except AttributeError:
            new.x0 = self.x0 + self.dx
        if n > 1:
            return new.diff(n-1, axis=axis)
        return new

    @wraps(numpy.ndarray.all, assigned=("__doc__",))
    def all(
        self,
        axis: int | None = None,
        out: numpy.ndarray | None = None,
    ) -> bool:
        """Return `True` if all values are `True` along the given axis."""
        return bool(numpy.all(self.value, axis=axis, out=out))

    # -- useful methods --------------

    def to_dqflag(
        self,
        name: str | None = None,
        minlen: int = 1,
        dtype: type | None = None,
        *,
        round: bool = False,  # noqa: A002
        label: str | None = None,
        description: str | None = None,
    ) -> DataQualityFlag:
        """Convert this series into a `~gwpy.segments.DataQualityFlag`.

        Each contiguous set of `True` values are grouped as a
        `~gwpy.segments.Segment` running from the GPS time the first
        found `True`, to the GPS time of the next `False` (or the end
        of the series)

        Parameters
        ----------
        name: `str`, optional
            Name of the segment.

        minlen : `int`, optional
            Minimum number of consecutive `True` values to identify as a
            `~gwpy.segments.Segment`. This is useful to ignore single
            bit flips, for example.

        dtype : `type`, `callable`
            Output segment entry type, can pass either a type for simple
            casting, or a callable function that accepts a float and returns
            another numeric type, defaults to the `dtype` of the time index.

        round : `bool`, optional
            Choose to round each `~gwpy.segments.Segment` to its
            inclusive integer boundaries.

        label : `str`, optional
            The :attr:`~gwpy.segments.DataQualityFlag.label` for the
            output flag.

        description : `str`, optional
            The :attr:`~gwpy.segments.DataQualityFlag.description` for the
            output flag.

        Returns
        -------
        dqflag : `~gwpy.segments.DataQualityFlag`
            A segment representation of this `StateTimeSeries`, the span
            defines the `known` segments, while the contiguous `True`
            sets defined each of the `active` segments.
        """
        from ..segments import DataQualityFlag

        # format dtype
        if dtype is None:
            dtype = self.t0.dtype
        if isinstance(dtype, numpy.dtype):  # use callable dtype
            dtype = dtype.type
        start = dtype(self.t0.value)
        dt = dtype(self.dt.value)

        # build segmentlists (can use simple objects since DQFlag converts)
        active = _bool_segments(self.value, start, dt, minlen=int(minlen))
        known = [tuple(map(dtype, self.span))]

        # build flag and return
        out = DataQualityFlag(
            name=name or self.name,
            active=active,
            known=known,
            label=label or self.name,
            description=description,
        )
        if round:
            return out.round()
        return out

    def to_lal(self) -> NoReturn:
        """Bogus function inherited from superclass, do not use."""
        msg = (
            "The to_lal method, inherited from the TimeSeries, cannot be used with the "
            "StateTimeSeries because LAL has no BooleanTimeSeries structure"
        )
        raise NotImplementedError(msg)

    @classmethod
    def from_nds2_buffer(
        cls,
        buffer: nds2.buffer,
        **metadata,
    ) -> Self:
        """Create a `StateTimeSeries` from an NDS2 buffer."""
        metadata.setdefault("unit", None)
        return super().from_nds2_buffer(buffer, **metadata)

    from_nds2_buffer.__doc__ = TimeSeriesBase.from_nds2_buffer.__doc__

    @overload  # type: ignore[override]
    def __getitem__(self, key: SupportsIndex) -> bool: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...
    def __getitem__(
        self,
        key: SupportsIndex | slice,
    ) -> bool | Self:
        """Return the value at the given index or slice."""
        if isinstance(key, float | int):
            return numpy.ndarray.__getitem__(self, key)
        return super().__getitem__(key)

    @wraps(TimeSeriesBase.tolist)
    def tolist(self) -> list:
        """Convert this `StateTimeSeries` to a list of boolean values."""
        return self.value.tolist()


# -- Bits ----------------------------

class Bits(list):
    """Definition of the bits in a `StateVector`.

    Parameters
    ----------
    bits : `list`
        list of bit names

    channel : `Channel`, `str`, optional
        data channel associated with this Bits

    epoch : `float`, optional
        defining GPS epoch for this `Bits`

    description : `dict`, optional
        (bit, desc) `dict` of longer descriptions for each bit
    """

    def __init__(
        self,
        bits: BitsInput,
        channel: Channel | str | None = None,
        epoch: Time | float | Quantity | None = None,
        description: dict[str, str] | None = None,
    ) -> None:
        """Initialize a new `Bits` object."""
        if isinstance(bits, dict):
            # dict of (index, bitname) pairs
            depth = max(map(int, bits.keys())) + 1
            super().__init__([None] * depth)
            for key, val in bits.items():
                self[int(key)] = val
        else:
            # list of names
            super().__init__((b or None) for i, b in enumerate(bits))

        # populate metadata
        if channel is not None:
            self.channel = channel
        if epoch is not None:
            self.epoch = epoch

        # populate descriptions
        if description is None:
            description = {}
        self.description = {}
        for i, bit in enumerate(self):
            if bit is None or bit in self.description:
                continue
            if channel:
                self.description[bit] = f"{self.channel} bit {i}"
            else:
                self.description[bit] = None

    @property
    def epoch(self) -> Time | None:
        """Starting GPS time epoch for these `Bits`.

        This attribute is recorded as a `~astropy.time.Time` object in the
        GPS format, allowing native conversion into other formats.

        See :mod:`~astropy.time` for details on the `Time` object.
        """
        try:
            return Time(self._epoch, format="gps")
        except AttributeError:
            return None

    @epoch.setter
    def epoch(self, epoch: Time | float | Quantity) -> None:
        if isinstance(epoch, Time):
            self._epoch = epoch.gps
        elif isinstance(epoch, units.Quantity):
            self._epoch = epoch.value
        else:
            self._epoch = float(epoch)

    @property
    def channel(self) -> Channel | None:
        """Data channel associated with these `Bits`."""
        try:
            return self._channel
        except AttributeError:
            return None

    @channel.setter
    def channel(self, chan: Channel | str) -> None:
        if isinstance(chan, Channel):
            self._channel = chan
        else:
            self._channel = Channel(chan)

    @property
    def description(self) -> dict[str, str | None]:
        """(name, desc) mapping of long bit descriptions."""
        return self._description

    @description.setter
    def description(
        self,
        desc: dict[str, str | None] | None,
    ) -> None:
        if desc is None:
            self._description = {}
        else:
            self._description = desc

    @description.deleter
    def description(self) -> None:
        self._description = {}

    def __repr__(self) -> str:
        """Return a string representation of this `Bits` object."""
        indent = " " * len(f"<{self.__class__.__name__}(")
        mask = (os.linesep + indent).join([
            f"{idx}: {bit!r}" for idx, bit in enumerate(self) if bit
        ])
        return os.linesep.join([
            f"<{self.__class__.__name__}({mask},",
            f"{indent}channel={self.channel!r},",
            f"{indent}epoch={self.epoch!r})>",
        ])

    def __str__(self) -> str:
        """Return a printable string representation of this `Bits` object."""
        indent = " " * len(f"{self.__class__.__name__}(")
        mask = (os.linesep + indent).join([
            f"{idx}: {bit}" for idx, bit in enumerate(self) if bit
        ])
        return os.linesep.join([
            f"{self.__class__.__name__}({mask},",
            f"{indent}channel={self.channel!s},",
            f"{indent}epoch={self.epoch!s})",
        ])

    @wraps(TimeSeriesBase.__array__)
    def __array__(self, dtype: DTypeLike = "U") -> numpy.ndarray:
        """Return a numpy array representation of this `Bits` object."""
        return numpy.array([b or "" for b in self], dtype=dtype)


# -- StateVector ---------------------

class StateVector(TimeSeriesBase):
    """Binary array representing good/bad state determinations of some data.

    Each binary bit represents a single boolean condition, with the
    definitions of all the bits stored in the `StateVector.bits`
    attribute.

    Parameters
    ----------
    value : array-like
        Input data array.

    bits : `Bits`, `list`, optional
        List of bits defining this `StateVector`.

    t0 : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    dt : `float`, `~astropy.units.Quantity`, optional
        Time between successive samples (seconds), can also be given inversely
        via `sample_rate`.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        The rate of samples per second (Hertz), can also be given inversely
        via `dt`.

    times : `array-like`
        The complete array of GPS times accompanying the data for this series.
        This argument takes precedence over `t0` and `dt` so should be given
        in place of these if relevant, not alongside.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional
        Choose to copy the input data to new memory.

    subok : `bool`, optional
        Allow passing of sub-classes by the array generator.

    Notes
    -----
    Key methods:

    .. autosummary::

        ~StateVector.fetch
        ~StateVector.read
        ~StateVector.write
        ~StateVector.to_dqflags
        ~StateVector.plot

    """

    _metadata_slots = (*TimeSeriesBase._metadata_slots, "bits")
    _print_slots = (*TimeSeriesBase._print_slots, "_bits")

    def __new__(
        cls,
        data: ArrayLike1D,
        bits: BitsInput | None = None,
        t0: SupportsToGps | None = None,
        dt: float | Quantity | None = None,
        sample_rate: float | Quantity | None = None,
        times: ArrayLike1D | None = None,
        channel: Channel | str | None = None,
        name: str | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new `StateVector`."""
        new = super().__new__(
            cls,
            data,
            t0=t0,
            dt=dt,
            sample_rate=sample_rate,
            times=times,
            channel=channel,
            name=name,
            **kwargs,
        )
        new.bits = bits
        return new

    # -- StateVector properties ------

    # -- bits
    @property
    def bits(self) -> Bits:
        """List of `Bits` for this `StateVector`.

        :type: `Bits`
        """
        try:
            return self._bits
        except AttributeError as exc:
            if self.dtype.kind in "iu":
                nbits = self.itemsize * 8
                self.bits = Bits(
                    [f"Bit {b}" for b in range(nbits)],
                    channel=self.channel,
                    epoch=self.epoch,
                )
                return self.bits

            if hasattr(self.channel, "bits"):
                self.bits = self.channel.bits  # type: ignore[union-attr]
                return self.bits

            msg = (
                "cannot determine bits for this StateVector, please set them "
                "explicitly via the 'bits' argument or attribute."
            )
            raise ValueError(msg) from exc

    @bits.setter
    def bits(
        self,
        mask: Bits | BitsInput | None,
    ) -> None:
        if mask is None:
            del self.bits
            return
        if not isinstance(mask, Bits):
            mask = Bits(
                mask,
                channel=self.channel,
                epoch=self.epoch,
            )
        self._bits = mask

    @bits.deleter
    def bits(self) -> None:
        with suppress(AttributeError):
            del self._bits

    # -- boolean
    @property
    def boolean(self) -> Array2D:
        """A 2-D boolean array representation of this `StateVector`."""
        try:
            return self._boolean
        except AttributeError:
            nbits = len(self.bits)
            boolean = numpy.zeros((self.size, nbits), dtype=bool)
            for i, sample in enumerate(self.value):
                boolean[i, :] = [int(sample) >> j & 1 for j in range(nbits)]
            self._boolean = Array2D(
                boolean,
                name=self.name,
                x0=self.x0,
                dx=self.dx,
                y0=0,
                dy=1,
            )
            return self.boolean

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(StateVectorRead)
    write = UnifiedReadWriteMethod(StateVectorWrite)
    get = UnifiedReadWriteMethod(StateVectorGet)

    # -- StateVector methods ---------

    def get_bit_series(
        self,
        bits: Iterable[int | str] | None = None,
    ) -> StateTimeSeriesDict:
        """Get the `StateTimeSeries` for each bit of this `StateVector`.

        Parameters
        ----------
        bits : `list`, optional
            A list of bit indices or bit names, defaults to all bits.

        Returns
        -------
        bitseries : `StateTimeSeriesDict`
            A `dict` of `StateTimeSeries`, one for each given bit.
        """
        if bits is None:
            bits = [b for b in self.bits if b]
        bindex = []
        try:
            for bit in bits:
                if isinstance(bit, int):
                    bindex.append((bit, self.bits[bit]))
                else:
                    bindex.append((self.bits.index(bit), bit))
        except (
            IndexError,
            ValueError,
        ) as exc:
            exc.args = (f"Bit {bit!r} not found in StateVector",)
            raise

        self._bitseries = StateTimeSeriesDict()
        for i, bit in bindex:
            self._bitseries[bit] = StateTimeSeries(
                self.value >> i & 1,
                name=bit,
                epoch=self.x0.value,
                channel=self.channel,
                sample_rate=self.sample_rate,
            )
        return self._bitseries

    def to_dqflags(
        self,
        bits: Iterable[int | str] | None = None,
        minlen: int = 1,
        dtype: type = float,
        *,
        round: bool = False,  # noqa: A002
    ) -> DataQualityDict:
        """Convert this `StateVector` into a `~gwpy.segments.DataQualityDict`.

        The `StateTimeSeries` for each bit is converted into a
        `~gwpy.segments.DataQualityFlag` with the bits combined into a dict.

        Parameters
        ----------
        bits : `list`, optional
            A list of bit indices or bit names to select,
            defaults to `~StateVector.bits`.

        minlen : `int`, optional
           Minimum number of consecutive `True` values to identify as a
           `Segment`. This is useful to ignore single bit flips, for example.

        dtype : `type`, optional
            Output segment entry type, can pass either a type for simple
            casting, or a callable function that accepts a float and returns
            another numeric type, defaults to `float`.

        round : `bool`, optional
            Choose to round each `Segment` to its inclusive integer boundaries.

        Returns
        -------
        DataQualityFlag list : `list`
            A list of `~gwpy.segments.flag.DataQualityFlag`
            representations for each bit in this `StateVector`.

        See Also
        --------
        StateTimeSeries.to_dqflag
            For details on the segment representation method for `StateVector` bits.
        """
        from ..segments import DataQualityDict

        out = DataQualityDict()
        bitseries = self.get_bit_series(bits=bits)
        for bit, sts in bitseries.items():
            name = str(bit)
            out[bit] = sts.to_dqflag(
                name=name,
                minlen=minlen,
                round=round,
                dtype=dtype,
                description=self.bits.description[name],
            )
        return out

    @classmethod
    def fetch(
        cls,
        channel: str | Channel,
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        bits: Bits | BitsInput | None = None,
        host: str | None = None,
        port: int | None = None,
        verbose: bool | str = False,
        connection: nds2.connection | None = None,
        verify: bool = False,
        pad: float | None = None,
        allow_tape: bool | None = None,
        scaled: bool | None = None,
        type: int | str | None = None,  # noqa: A002
        dtype: int | str | None = None,
    ) -> Self:
        """Fetch data from NDS into a `StateVector`.

        Parameters
        ----------
        channel : `str`, `~gwpy.detector.Channel`
            The name (or representation) of the data channel to fetch.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        bits : `Bits`, `list`, optional
            Definition of bits for this `StateVector`.

        host : `str`, optional
            URL of NDS server to use, if blank will try any server
            (in a relatively sensible order) to get the data

            One of ``connection`` or ``host`` must be given.

        port : `int`, optional
            Port number for NDS server query, must be given with `host`.

        verify : `bool`, optional
            Check channels exist in database before asking for data.
            Default is `True`.

        pad : `float`, optional
            Value with which to fill gaps in the source data.
            By default gaps will result in a `ValueError`.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        connection : `nds2.connection`, optional
            Open NDS connection to use.
            Default is to open a new connection using ``host`` and ``port``
            arguments.

            One of ``connection`` or ``host`` must be given.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        allow_tape : `bool`, optional
            Allow data access from slow tapes.
            If ``host`` or ``connection`` is given, the default is to do
            whatever the server default is, otherwise servers will be searched
            with ``allow_tape=False`` first, then ``allow_tape=True`` if that
            fails.

        type : `int`, `str`, optional
            NDS2 channel type integer or string name to match.
            Default is to search for any channel type.

        dtype : `numpy.dtype`, `str`, `type`, or `dict`, optional
            NDS2 data type to match.
            Default is to search for any data type.
        """
        new = super().fetch(
            channel,
            start,
            end,
            host=host,
            port=port,
            verbose=verbose,
            connection=connection,
            verify=verify,
            pad=pad,
            scaled=scaled,
            allow_tape=allow_tape,
            type=type,
            dtype=dtype,
        )
        if bits:
            new.bits = bits
        return new

    def plot(  # type: ignore[override]
        self,
        format: Literal["timeseries", "segments"] = "segments",  # noqa: A002
        bits: Iterable[int | str] | None = None,
        **kwargs,
    ) -> Plot:
        """Plot the data for this `StateVector`.

        Parameters
        ----------
        format : `str`, optional
            The type of plot to make, either 'segments' to plot the
            SegmentList for each bit, or 'timeseries' to plot the raw
            data for this `StateVector`.

        bits : `list`, optional
            A list of bit indices or bit names, defaults to
            `~StateVector.bits`. This argument is ignored if ``format`` is
            not ``'segments'``.

        kwargs
            Other keyword arguments to be passed to either
            `~gwpy.plot.SegmentAxes.plot` or `~gwpy.plot.Axes.plot`, depending
            on ``format``.

        Returns
        -------
        plot : `~gwpy.plot.Plot`
            Output plot object.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the figure.

        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the axes.

        gwpy.plot.SegmentAxes.plot_flag
            For documentation of keyword arguments used in rendering each
            statevector flag.
        """
        if format == "timeseries":
            return super().plot(**kwargs)

        if format == "segments":
            from ..plot import Plot

            kwargs.setdefault("xscale", "auto-gps")
            return Plot(
                *self.to_dqflags(bits=bits).values(),
                projection="segments",
                **kwargs,
            )

        msg = "'format' argument must be one of: 'timeseries' or 'segments'"
        raise ValueError(msg)

    def resample(self, rate: float | Quantity) -> StateVector:
        """Resample this `StateVector` to a new rate.

        Because of the nature of a state-vector, downsampling is done
        by taking the logical 'and' of all original samples in each new
        sampling interval, while upsampling is achieved by repeating
        samples.

        Parameters
        ----------
        rate : `float`
            Rate to which to resample this `StateVector`, must be a
            divisor of the original sample rate (when downsampling)
            or a multiple of the original (when upsampling).

        Returns
        -------
        vector : `StateVector`
            Resampled version of the input `StateVector`.
        """
        rate1 = self.sample_rate.value
        if isinstance(rate, units.Quantity):
            rate2 = rate.value
        else:
            rate2 = float(rate)

        # upsample
        if (rate2 / rate1).is_integer():
            msg = "StateVector upsampling has not been implemented yet, sorry."
            raise NotImplementedError(msg)

        # downsample
        if (rate1 / rate2).is_integer():
            factor = int(rate1 / rate2)
            # reshape incoming data to one column per new sample
            newsize = int(self.size / factor)
            old = self.value.reshape((newsize, self.size // newsize))
            # work out number of bits
            if self.bits:
                nbits = len(self.bits)
            else:
                max_ = self.value.max()
                nbits = ceil(log2(max_)) if max_ else 1
            bits = range(nbits)
            # construct an iterator over the columns of the old array
            itr = numpy.nditer(
                [old, None],
                flags=["external_loop", "reduce_ok"],
                op_axes=[None, [0, -1]],  # type: ignore[list-item]
                op_flags=[["readonly"], ["readwrite", "allocate"]],
            )
            dtype = self.dtype
            type_ = self.dtype.type
            # for each new sample, each bit is logical AND of old samples
            # bit is ON,
            for x, y in itr:
                y[...] = numpy.sum(
                    [type_((x >> bit & 1).all() * (2 ** bit)) for bit in bits],
                    dtype=self.dtype,
                )
            new = StateVector(itr.operands[1], dtype=dtype)
            new.__metadata_finalize__(self)
            new._unit = self.unit
            new.sample_rate = rate2
            return new

        # error for non-integer resampling factors
        if rate1 < rate2:
            msg = (
                "New sample rate must be multiple of input series rate if "
                "upsampling a StateVector"
            )
            raise ValueError(msg)
        msg = (
            "New sample rate must be divisor of input series rate if "
            "downsampling a StateVector"
        )
        raise ValueError(msg)


@as_series_dict_class(StateTimeSeries)
class StateTimeSeriesDict(TimeSeriesBaseDict):
    """Dictionary of `StateTimeSeries` objects."""

    __doc__ = TimeSeriesBaseDict.__doc__.replace("TimeSeriesBase", "StateTimeSeries")  # type: ignore[union-attr]
    EntryClass = StateTimeSeries


@as_series_dict_class(StateVector)
class StateVectorDict(TimeSeriesBaseDict):
    """Dictionary of `StateVector` objects."""

    __doc__ = TimeSeriesBaseDict.__doc__.replace("TimeSeriesBase", "StateVector")  # type: ignore[union-attr]
    EntryClass = StateVector

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(StateVectorDictRead)
    write = UnifiedReadWriteMethod(StateVectorDictWrite)
    get = UnifiedReadWriteMethod(StateVectorDictGet)


class StateVectorList(TimeSeriesBaseList):
    """List of `StateVector` objects."""

    __doc__ = TimeSeriesBaseList.__doc__.replace("TimeSeriesBase", "StateVector")  # type: ignore[union-attr]
    EntryClass = StateVector



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.spectrum`."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
)

import pytest
from astropy.time import Time
from numpy.random import default_rng

from ...frequencyseries import FrequencySeries
from .. import SpectrumProduct
from .base import _TestFrequencyDomainProduct

if TYPE_CHECKING:
    from typing import ClassVar

SpectrumProductType = TypeVar("SpectrumProductType", bound=SpectrumProduct)


class TestSpectrumProduct(
    _TestFrequencyDomainProduct[SpectrumProductType],
    Generic[SpectrumProductType],
):
    """Tests for `gwpy.cli.SpectrumProduct`."""

    TEST_CLASS: ClassVar[type[SpectrumProduct]] = SpectrumProduct
    ACTION = "spectrum"

    @pytest.fixture
    @classmethod
    def dataprod(cls, prod: SpectrumProductType) -> SpectrumProductType:
        """Return a `TestFrequencyDomainProduct` with data."""
        cls._prod_add_data(prod)
        fftlength = prod.args.secpfft
        for i, ts in enumerate(prod.timeseries):
            nsamp = int(fftlength * 512 / 2.) + 1
            rng = default_rng(i)
            fs = FrequencySeries(
                rng.random(nsamp),
                x0=0,
                dx=1/fftlength,
                channel=ts.channel,
                name=ts.name,
            )
            prod.spectra.append(fs)
        return prod

    def test_get_title(self, prod: SpectrumProductType):
        """Test `SpectrumProduct.get_title`."""
        epoch = prod.start_list[0]
        utc = Time(epoch, format="gps", scale="utc").iso
        t = ", ".join([
            f"{utc} | {epoch} ({prod.duration})",
            f"fftlength={prod.args.secpfft}",
            f"overlap={prod.args.overlap}",
        ])
        assert prod.get_title() == t

    def test_get_suptitle(self, prod: SpectrumProductType):
        """Test `SpectrumProduct.get_suptitle`."""
        assert prod.get_suptitle() == f"Spectrum: {prod.chan_list[0]}"



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for timeseries module."""

from __future__ import annotations

import logging
import sys
from contextlib import nullcontext
from itertools import (
    chain,
    product,
)
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    TypeVar,
)
from unittest import mock

import numpy
import pytest
from astropy import units
from numpy import testing as nptest
from requests.exceptions import HTTPError
from scipy import signal
from scipy.signal.windows import tukey

from ...frequencyseries import FrequencySeries, SpectralVariance
from ...io.gwf import get_backend as get_gwf_backend
from ...segments import DataQualityFlag, Segment, SegmentList
from ...signal import filter_design
from ...spectrogram import Spectrogram
from ...table import EventTable
from ...testing import mocks, utils
from ...testing.errors import (
    pytest_skip_flaky_network,
    pytest_skip_network_error,
)
from ...time import LIGOTimeGPS
from ...types import Index
from .. import StateTimeSeries, TimeSeries, TimeSeriesDict, TimeSeriesList
from ..io import gwf as tsio_gwf
from .test_core import (
    TestTimeSeriesBase as _TestTimeSeriesBase,
    TestTimeSeriesBaseDict as _TestTimeSeriesBaseDict,
    TestTimeSeriesBaseList as _TestTimeSeriesBaseList,
)

if TYPE_CHECKING:
    from typing import ClassVar

TimeSeriesType = TypeVar("TimeSeriesType", bound=TimeSeries)

try:
    get_gwf_backend(
        "gwpy.timeseries.io.gwf",
        backends=tsio_gwf.BACKENDS,
    )
except ImportError:
    HAVE_GWF_BACKEND = False
else:
    HAVE_GWF_BACKEND = True

RNG = numpy.random.default_rng(0)

# remote URL for test .gwf file
TEST_HDF5_URL = (
    "https://gitlab.com/gwpy/gwpy/-/raw/v3.0.10/"
    + Path(utils.TEST_HDF5_FILE).relative_to(
        utils.TEST_DATA_PATH.parent.parent.parent,
    ).as_posix()
)

SKIP_GWF_BACKEND = pytest.mark.skipif(
    not HAVE_GWF_BACKEND,
    reason="no GWF backend",
)
GWF_BACKENDS = [
    pytest.param(None, marks=SKIP_GWF_BACKEND),
    pytest.param("lalframe", marks=pytest.mark.requires("lalframe")),
    pytest.param("framecpp", marks=pytest.mark.requires("LDAStools.frameCPP")),
    pytest.param("framel", marks=pytest.mark.requires("framel")),
]

LIVETIME = DataQualityFlag(
    name="X1:TEST-FLAG:1",
    active=SegmentList([
        Segment(0, 32),
        Segment(34, 34.5),
    ]),
    known=SegmentList([Segment(0, 64)]),
    isgood=True,
)

GWOSC_DATAFIND_SERVER = "datafind.gwosc.org"

# Description of GW150914 for tests
# This is used in various tests
GWOSC_GW150914_IFO = "L1"
GWOSC_GW150914_CHANNEL = "L1:GWOSC-16KHZ_R1_STRAIN"
NDS2_GW150914_CHANNEL = "L1:DCS-CALIB_STRAIN_C02"
GWOSC_GW150914_FRAMETYPE = "L1_LOSC_16_V1"
GWOSC_GW150914 = 1126259462
GWOSC_GW150914_SEGMENT: Segment[float] = Segment(GWOSC_GW150914 - 2, GWOSC_GW150914 + 2)
GWOSC_GW150914_SEGMENT_32: Segment[float] = Segment(
    GWOSC_GW150914 - 16,
    GWOSC_GW150914 + 16,
)
GWOSC_GW150914_DQ_NAME = {
    "hdf5": "L1:DQmask",
    "gwf": "L1:GWOSC-4KHZ_R1_DQMASK",
}
GWOSC_GW150914_DQ_BITS = {
    "hdf5": [
        "Passes DATA test",
        "Passes CBC_CAT1 test",
        "Passes CBC_CAT2 test",
        "Passes CBC_CAT3 test",
        "Passes BURST_CAT1 test",
        "Passes BURST_CAT2 test",
        "Passes BURST_CAT3 test",
    ],
    "gwf": [
        "DATA",
        "CBC_CAT1",
        "CBC_CAT2",
        "CBC_CAT3",
        "BURST_CAT1",
        "BURST_CAT2",
        "BURST_CAT3",
    ],
}
GWOSC_GW150914_INJ_NAME = {
    "hdf5": "L1:Injmask",
    "gwf": "L1:GWOSC-4KHZ_R1_INJMASK",
}
GWOSC_GW150914_INJ_BITS = {
    "hdf5": [
        "Passes NO_CBC_HW_INJ test",
        "Passes NO_BURST_HW_INJ test",
        "Passes NO_DETCHAR_HW_INJ test",
        "Passes NO_CW_HW_INJ test",
        "Passes NO_STOCH_HW_INJ test",
    ],
    "gwf": [
        "NO_CBC_HW_INJ",
        "NO_BURST_HW_INJ",
        "NO_DETCHAR_HW_INJ",
        "NO_CW_HW_INJ",
        "NO_STOCH_HW_INJ",
    ],
}

# Description of GW190814
# This is used in various tests in particular to test GWOSC new file format
# (but not for NDS and DataFind)
GWOSC_GW190814_IFO = "L1"
GWOSC_GW190814_CHANNEL = "L1:GWOSC-16KHZ_R1_STRAIN"
GWOSC_GW190814 = 1249852257
GWOSC_GW190814_SEGMENT: Segment[float] = Segment(GWOSC_GW190814 - 2, GWOSC_GW190814 + 2)
GWOSC_GW190814_SEGMENT_32: Segment[float] = Segment(
    GWOSC_GW190814 - 16,
    GWOSC_GW190814 + 16,
)
GWOSC_GW190814_DQ_NAME =  {
    "hdf5": "L1:GWOSC-4KHZ_R1_DQMASK",
    "gwf": "L1:GWOSC-4KHZ_R1_DQMASK",
}
GWOSC_GW190814_DQ_BITS = GWOSC_GW150914_DQ_BITS
GWOSC_GW190814_INJ_NAME = {
    "hdf5": "L1:GWOSC-4KHZ_R1_INJMASK",
    "gwf": "L1:GWOSC-4KHZ_R1_INJMASK",
}
GWOSC_GW190814_INJ_BITS = GWOSC_GW150914_INJ_BITS


__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _gwosc_pelican(func):
    """Decorate ``func`` with all necessary GWOSC/Pelican-related decorators."""
    for dec in (
        pytest.mark.requires("lalframe"),
        pytest.mark.requires("requests_pelican"),
        pytest_skip_flaky_network,
        mock.patch.dict(
            "os.environ",
            {"GWDATAFIND_SERVER": GWOSC_DATAFIND_SERVER},
        ),
    ):
        func = dec(func)
    return func


class TestTimeSeries(_TestTimeSeriesBase[TimeSeriesType]):
    """Test `TimeSeries`."""

    TEST_CLASS: type[TimeSeriesType] = TimeSeries

    # -- fixtures --------------------

    @pytest.fixture(scope="class")
    @pytest_skip_network_error
    def gw150914(self) -> TimeSeriesType:
        """TimeSeries containing GW150914 data from GWOSC at 4096 Hz."""
        return self.TEST_CLASS.get(
            GWOSC_GW150914_IFO,
            *GWOSC_GW150914_SEGMENT,
            sample_rate=4096,
        )

    @pytest.fixture(scope="class")
    @pytest_skip_network_error
    def gw150914_16384(self) -> TimeSeriesType:
        """TimeSeries containing GW150914 data from GWOSC at 16384 Hz."""
        return self.TEST_CLASS.get(
            GWOSC_GW150914_IFO,
            GWOSC_GW150914_SEGMENT.start,
            GWOSC_GW150914_SEGMENT.end,
            sample_rate=16384,
        )

    @pytest.fixture(scope="class")
    @pytest_skip_network_error
    def gw150914_h1_32(self) -> TimeSeriesType:
        """TimeSeries containing 32-seconds of H1 GW150914 data from GWOSC."""
        return self.TEST_CLASS.get(
            "H1",
            GWOSC_GW150914_SEGMENT_32.start,
            GWOSC_GW150914_SEGMENT_32.end,
            source="gwosc",
            sample_rate=4096,
        )

    @pytest.fixture(scope="class")
    @pytest_skip_network_error
    def gw150914_l1_32(self) -> TimeSeriesType:
        """TimeSeries containing 32-seconds of L1 GW150914 data from GWOSC."""
        return self.TEST_CLASS.get(
            "L1",
            GWOSC_GW150914_SEGMENT_32.start,
            GWOSC_GW150914_SEGMENT_32.end,
            source="gwosc",
            sample_rate=4096,
        )

    # -- test class functionality ----

    def test_ligotimegps(self):
        """Test that TimeSeries.t0 and x0 can be set with LIGOTimeGPS."""
        # test that LIGOTimeGPS works
        array = self.create(t0=LIGOTimeGPS(0))
        assert array.t0.value == 0
        array.t0 = LIGOTimeGPS(10)
        assert array.t0.value == 10
        array.x0 = LIGOTimeGPS(1000000000)
        assert array.t0.value == 1000000000

        # check epoch access
        array.epoch = LIGOTimeGPS(10)
        assert array.t0.value == 10

    def test_epoch(self):  # type: ignore[override]
        """Test `TimeSeries.epoch`."""
        array = self.create()
        assert array.epoch.gps == array.x0.value  # type: ignore[union-attr]

    # -- test I/O --------------------

    def test_read_cache(self, tmp_path):
        """Test that `TimeSeries.read` handles caches well."""
        # create an array and write it to a file
        array = self.create(t0=0, dt=10/self.data.size, name="TEST1")
        file1 = tmp_path / "X-data-0-10.h5"
        array.write(file1)

        # create a second array, but don't write it
        file2 = tmp_path / "X-data-10-10.h5"

        # create a cache that talks about both files
        cache = [file1, file2]

        # check that when reading _and_ asking for times covered by the
        # first file, the cache is appropriately filtered
        data = self.TEST_CLASS.read(
            cache,
            "TEST1",
            format="hdf5",
            start=0,
            end=10,
        )
        utils.assert_quantity_sub_equal(array, data)

    def test_read_start_end_args(self, tmp_path):
        """Test that `TimeSeries.read` handles start/end as positionals."""
        # Create an array and write it to a file
        array = self.create(t0=0, dt=10/self.data.size, name="TEST1")
        h5file = tmp_path / "X-data-0-10.h5"
        array.write(h5file)

        # Try and read (half) the file passing start/end as positionals
        data = self.TEST_CLASS.read(
            h5file,
            "TEST1",
            0,
            5,
            format="hdf5",
        )
        half = array.size // 2
        utils.assert_quantity_sub_equal(array[:half], data)

    @pytest.mark.parametrize("fmt", ["txt", "csv"])
    def test_read_write_ascii(self, array, fmt):
        """Test reading and writing ASCII files."""
        utils.test_read_write(
            array,
            fmt,
            assert_equal=utils.assert_quantity_sub_equal,
            assert_kw={"exclude": ["name", "channel", "unit"]},
        )

    def test_read_ascii_header(self, tmpdir):
        """Check that ASCII files with headers are read without extra options.

        [regression: https://gitlab.com/gwpy/gwpy/-/issues/1473]
        """
        txt = tmpdir / "text.txt"
        txt.write_text(
            "# time (s)\tdata (strain)\n0\t1\n1\t2\n2\t3",
            encoding="utf-8",
        )
        data = self.TEST_CLASS.read(txt, format="txt")
        utils.assert_array_equal(data.times, Index((0, 1, 2), unit="s"))
        utils.assert_array_equal(data.value, (1, 2, 3))

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_read_write_gwf(self, tmp_path, backend):
        """Test reading and writing GWF files."""
        array = self.create(name="TEST")

        # test basic write/read
        try:
            utils.test_read_write(
                array,
                "gwf",
                extension="gwf",
                read_args=[array.name],
                read_kw={"backend": backend},
                write_kw={"backend": backend},
                assert_equal=utils.assert_quantity_sub_equal,
                assert_kw={"exclude": ["channel"]},
            )
        except ImportError as e:  # pragma: no-cover
            pytest.skip(str(e))

        # test read keyword arguments
        tmp = tmp_path / "test.gwf"
        array.write(tmp, format="gwf", backend=backend)

        def read_(**kwargs):
            return type(array).read(
                tmp,
                array.name,
                format="gwf",
                backend=backend,
                **kwargs,
            )

        # test start and end
        start, end = array.span.contract(10)
        t = read_(start=start, end=end)
        utils.assert_quantity_sub_equal(
            t, array.crop(start, end),
            exclude=["channel"],
        )
        assert t.span == (start, end)

        # test start only
        t = read_(start=start)
        utils.assert_quantity_sub_equal(
            t,
            array.crop(start=start),
            exclude=["channel"],
        )

        # test end only
        t = read_(end=end)
        utils.assert_quantity_sub_equal(
            t,
            array.crop(end=end),
            exclude=["channel"],
        )

    @SKIP_GWF_BACKEND
    def test_write_gwf_overwrite(self, tmp_path):
        """Test that writing GWF with overwrite works."""
        tmp = tmp_path / "test.gwf"

        # Create a file (that mustn't already exist)
        array = self.create(name="TEST")
        array.write(tmp, format="gwf", overwrite=False)

        # check that we can't write again without overwrite
        with pytest.raises(
            OSError,
            match="File exists",
        ):
            array.write(tmp, format="gwf", overwrite=False)

        # check that overwrite=True is default and works
        array.write(tmp, format="gwf")

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_read_gwf_end_error(self, backend):
        """Test that reading past the end of available data fails."""
        with pytest.raises(
            ValueError,
            match=r"(cannot read data|failed to read)",
        ):
            self.TEST_CLASS.read(
                utils.TEST_GWF_FILE,
                "L1:LDAS-STRAIN",
                format="gwf",
                backend=backend,
                start=utils.TEST_GWF_SPAN[1],
            )

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_read_gwf_negative_duration_error(self, backend):
        """Test that reading a negative duration fails."""
        with pytest.raises(
            ValueError,
            match=r"(cannot read data|failed to read)",
        ):
            self.TEST_CLASS.read(
                utils.TEST_GWF_FILE,
                "L1:LDAS-STRAIN",
                format="gwf",
                backend=backend,
                end=utils.TEST_GWF_SPAN[0]-1,
            )

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    @pytest.mark.parametrize("parallel", [1, 2])
    def test_read_write_gwf_multiple(self, tmp_path, backend, parallel):
        """Check that each GWF backend can read a series of files.

        Regression: https://gitlab.com/gwpy/gwpy/-/issues/1486
        """
        a1 = self.create(name="TEST")
        a2 = self.create(name="TEST", t0=a1.span[1], dt=a1.dx)

        tmp1 = tmp_path / "test1.gwf"
        tmp2 = tmp_path / "test2.gwf"
        a1.write(tmp1, format="gwf", backend=backend)
        a2.write(tmp2, format="gwf", backend=backend)
        cache = [tmp1, tmp2]

        comb = self.TEST_CLASS.read(
            cache,
            "TEST",
            start=a1.span[0],
            end=a2.span[1],
            format="gwf",
            backend=backend,
            parallel=parallel,
        )
        utils.assert_quantity_sub_equal(
            comb,
            a1.append(a2, inplace=False),
            exclude=["channel"],
        )

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_read_write_gwf_name_error(self, tmp_path, backend, gw150914):
        """Test that reading GWF handles missing/wrong channel names."""
        tmp = tmp_path / "test.gwf"
        gw150914.write(tmp, format="gwf", backend=backend)

        # wrong channel (framel on windows gives a slightly different error)
        with pytest.raises(
            ValueError,
            match=r"({})".format("|".join((  # noqa: FLY002
                "^channel 'another channel' not found",
                "vector not found: another channel",
            ))),
        ):
            self.TEST_CLASS.read(
                tmp,
                "another channel",
                format="gwf",
                backend=backend,
            )

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_read_write_gwf_interval_error(self, tmp_path, backend, gw150914):
        """Test that reading GWF handles out-of-bounds times."""
        tmp = tmp_path / "test.gwf"
        gw150914.write(tmp, format="gwf", backend=backend)

        # wrong times (error message is different for each backend)
        with pytest.raises(
            ValueError,
            match=r"(cannot read data|failed to read)",
        ):
            self.TEST_CLASS.read(
                tmp,
                gw150914.name,
                start=gw150914.span[0]-1,
                end=gw150914.span[0],
                format="gwf",
                backend=backend,
            )

    @pytest.mark.requires("lalframe")  # for get_channel_type
    @pytest.mark.parametrize("ctype", ["adc", "proc", "sim", None])
    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_write_gwf_type(self, gw150914, tmp_path, backend, ctype):
        """Test writing GWF for each FrData type."""
        from ...io.gwf import get_channel_type

        expected_ctype = ctype or "proc"

        # write the file
        tmp = tmp_path / "test.gwf"
        gw150914.write(tmp, type=ctype, format="gwf", backend=backend)

        # assert that the type is correct
        assert get_channel_type(gw150914.name, tmp) == expected_ctype

        # read the file and check that the data match
        new = type(gw150914).read(
            tmp,
            gw150914.name,
            format="gwf",
            backend=backend,
        )

        # epoch seems to mismatch at O(1e-12), which is unfortunate
        utils.assert_quantity_sub_equal(
            gw150914,
            new,
            exclude=("channel", "x0"),
        )

    @pytest.mark.parametrize("backend", GWF_BACKENDS)
    def test_write_gwf_channel_name(self, tmp_path, backend):
        """Test that writing GWF uses `channel` when `name` is missing."""
        # create data
        array = self.create(channel="data")
        assert not array.name

        # write data
        tmp = tmp_path / "test.gwf"
        array.write(tmp, format="gwf", backend=backend)

        # read it back check
        array2 = type(array).read(tmp, str(array.channel), format="gwf")
        assert array2.name == str(array.channel)
        utils.assert_quantity_sub_equal(
            array,
            array2,
            exclude=("name", "channel"),
        )

    @pytest.mark.parametrize("ext", ["hdf5", "h5"])
    @pytest.mark.parametrize("channel", [
        None,
        "test",
        "X1:TEST-CHANNEL",
    ])
    def test_read_write_hdf5(self, tmp_path, ext, channel):
        """Test reading and writing HDF5 files."""
        array = self.create()
        array.channel = channel

        tmp = tmp_path / f"test.{ext}"
        # check array with no name fails
        with pytest.raises(
            ValueError,
            match=r"^Cannot determine HDF5 path",
        ):
            array.write(tmp, overwrite=True)
        array.name = "TEST"

        # write array (with auto-identify)
        array.write(tmp, overwrite=True)

        # check reading gives the same data (with/without auto-identify)
        ts = type(array).read(tmp, format="hdf5")
        utils.assert_quantity_sub_equal(array, ts)
        ts = type(array).read(tmp)
        utils.assert_quantity_sub_equal(array, ts)

        # check that we can't then write the same data again
        with pytest.raises(
            OSError,
            match="File exists",
        ):
            array.write(tmp)
        with pytest.raises((IOError, OSError, RuntimeError, ValueError)):
            array.write(tmp, append=True)

        # check reading with start/end works
        start, end = array.span.contract(25)
        t = type(array).read(tmp, start=start, end=end)
        utils.assert_quantity_sub_equal(t, array.crop(start, end))

    def test_read_write_wav(self):
        """Test reading and writing WAV files."""
        array = self.create(dtype="float32")
        utils.test_read_write(
            array, "wav", read_kw={"mmap": True}, write_kw={"scale": 1},
            assert_equal=utils.assert_quantity_sub_equal,
            assert_kw={"exclude": ["unit", "name", "channel", "x0"]})

    @pytest.mark.parametrize(("pre", "post"), [
        pytest.param(None, None, id="none"),
        pytest.param(0, 0, id="zero"),
        pytest.param(None, 1, id="right"),
        pytest.param(1, None, id="left"),
        pytest.param(1, 1, id="both"),
    ])
    def test_read_pad(self, pre, post):
        """Check that `TimeSeries.read` with `pad` works as expected."""
        a = self.TEST_CLASS.read(
            utils.TEST_HDF5_FILE,
            "H1:LDAS-STRAIN",
        )
        start = None if pre is None else a.span[0] - pre
        end = None if post is None else a.span[1] + post
        b = self.TEST_CLASS.read(
            utils.TEST_HDF5_FILE,
            "H1:LDAS-STRAIN",
            pad=0.,
            start=start,
            end=end,
        )
        pres = 0 if not pre else int(pre * a.sample_rate.value)
        posts = 0 if not post else int(post * a.sample_rate.value)
        utils.assert_quantity_sub_equal(
            a.pad(
                (pres, posts),
                mode="constant",
                constant_values=(0,),
            ),
            b,
        )

    def test_read_pad_raise(self):
        """Check that `TimeSeries.read` with `gap='raise'` raises appropriately.

        [regression: https://gitlab.com/gwpy/gwpy/-/issues/1211]
        """
        from gwpy.io.cache import file_segment
        span = file_segment(utils.TEST_HDF5_FILE)
        with pytest.raises(
            ValueError,
            match="does not cover requested interval",
        ):
            self.TEST_CLASS.read(
                utils.TEST_HDF5_FILE,
                "H1:LDAS-STRAIN",
                pad=0.,
                start=span[0],
                end=span[1]+1.,
                gap="raise",
            )

    @pytest_skip_flaky_network
    def test_read_remote(self):
        """Test that reading directly from a remote URI works."""
        local = self.TEST_CLASS.read(
            utils.TEST_HDF5_FILE,
            "H1:LDAS-STRAIN",
        )
        remote = self.TEST_CLASS.read(
            TEST_HDF5_URL,
            "H1:LDAS-STRAIN",
            cache=False,
        )
        utils.assert_quantity_sub_equal(local, remote)

    @pytest.mark.requires("nds2")
    def test_from_nds2_buffer_dynamic_scaled(self):
        """Test `TimeSeries.from_nds2_buffer` with scaled data."""
        # build fake buffer for LIGO channel
        nds_buffer = mocks.nds2_buffer(
            "H1:TEST",
            self.data,
            1000000000,
            self.data.shape[0],
            "m",
            name="test",
            slope=2,
            offset=1,
        )

        # check scaling defaults to off
        utils.assert_array_equal(
            self.TEST_CLASS.from_nds2_buffer(nds_buffer).value,
            nds_buffer.data,
        )
        utils.assert_array_equal(
            self.TEST_CLASS.from_nds2_buffer(nds_buffer, scaled=False).value,
            nds_buffer.data,
        )
        utils.assert_array_equal(
            self.TEST_CLASS.from_nds2_buffer(nds_buffer, scaled=True).value,
            nds_buffer.data * 2 + 1,
        )

    # -- test remote data access -----

    @pytest.mark.parametrize("fmt", [
        "hdf5",
        pytest.param("gwf", marks=pytest.mark.requires("lalframe")),
    ])
    @pytest_skip_flaky_network
    def test_fetch_open_data(self, gw150914, fmt):
        """Test `TimeSeries.fetch_open_data`."""
        ts = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW150914_IFO,
            GWOSC_GW150914_SEGMENT.start,
            GWOSC_GW150914_SEGMENT.end,
            format=fmt,
            timeout=60,
        )
        utils.assert_quantity_sub_equal(
            ts,
            gw150914,
            exclude=["name", "unit", "channel"],
        )

        # try again with 16384 Hz data
        ts = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW150914_IFO,
            GWOSC_GW150914_SEGMENT.start,
            GWOSC_GW150914_SEGMENT.end,
            format=fmt,
            sample_rate=16384,
            timeout=60,
        )
        assert ts.sample_rate == 16384 * units.Hz

    @pytest_skip_flaky_network
    def test_fetch_open_data_error(self):
        """Test that TimeSeries.fetch_open_data propagates errors."""
        with pytest.RaisesGroup(
            pytest.RaisesExc(ValueError, match="Cannot find a GWOSC dataset"),
            match="failed to get data",
        ):
            self.TEST_CLASS.fetch_open_data(
                GWOSC_GW150914_IFO,
                0,
                1,
            )

    @pytest.mark.requires("nds2")
    @pytest.mark.parametrize("protocol", [
        pytest.param(1, id="nds1"),
        pytest.param(2, id="nds2"),
    ])
    def test_fetch(self, nds2_connection, protocol):
        """Test `TimeSeries.fetch`."""
        # set protocol
        nds2_connection.get_protocol.return_value = protocol

        # get expected result
        expected = self.TEST_CLASS.from_nds2_buffer(nds2_connection._buffers[0])

        # execute fetch()
        ts = self.TEST_CLASS.fetch(
            "X1:test",
            1000000000,
            1000000004,
        )
        utils.assert_quantity_sub_equal(ts, expected)

    @pytest.mark.requires("nds2")
    def test_fetch_connection(self, nds2_connection):
        """Test `TimeSeries.fetch(..., connection=<>)`."""
        expected = self.TEST_CLASS.from_nds2_buffer(nds2_connection._buffers[0])
        ts = self.TEST_CLASS.fetch(
            "X1:test",
            1000000000,
            1000000004,
            connection=nds2_connection,
        )
        utils.assert_quantity_sub_equal(ts, expected)

    @pytest.mark.requires("nds2")
    @pytest.mark.parametrize("protocol", [
        pytest.param(1, id="nds1"),
        pytest.param(2, id="nds2"),
    ])
    def test_fetch_pad(self, caplog, nds2_connection, protocol):
        """Test `TimeSeries.fetch(..., pad=...)`."""
        # set protocol
        nds2_connection.get_protocol.return_value = protocol

        # get expected result
        expected = self.TEST_CLASS.from_nds2_buffer(nds2_connection._buffers[0])

        with caplog.at_level(logging.DEBUG, logger="gwpy.timeseries.io.nds2"):
            # check padding works
            ts = self.TEST_CLASS.fetch(
                "X1:test",
                *expected.span.protract(2),
                pad=-100.,
                connection=nds2_connection,
            )

        assert ts.span == expected.span.protract(2)
        nptest.assert_array_equal(
            ts.value,
            numpy.concatenate((
                numpy.ones(int(2 * ts.sample_rate.value)) * -100.,
                expected.value,
                numpy.ones(int(2 * ts.sample_rate.value)) * -100.,
            )),
        )

        # check that the logger emitted warnings about the padding
        if protocol == 2 and sys.version_info >= (3, 12):
            for msg in (
                "[nds.test.gwpy] Availability check complete, "
                "found 1 viable segments of data with 66.67% coverage",
                "[nds.test.gwpy] Gaps will be padded with -100.0",
            ):
                assert (
                    "gwpy.timeseries.io.nds2",
                    logging.DEBUG,
                    msg,
                ) in caplog.record_tuples

    @pytest.mark.requires("nds2")
    def test_fetch_empty_iterate_error(self, nds2_connection):
        """Test `TimeSeries.fetch()` handling of no data."""
        # patch find_channels() to return the channel, even though
        # iterate() won't return any data
        def find_channels(name, *args, **kwargs):  # noqa: ARG001
            return [mocks.nds2_channel(name, 128, "")]

        nds2_connection.find_channels = find_channels
        nds2_connection._buffers = []

        # check that we get the right error
        with pytest.RaisesGroup(
            pytest.RaisesExc(RuntimeError, match="no data received from nds.test.gwpy"),
            match="failed to get data",
        ):
            self.TEST_CLASS.fetch("X1:test-missing", 0, 1, host="nds.gwpy")

    @_gwosc_pelican
    @pytest.mark.parametrize("kwargs", [
        pytest.param({"verbose": True}, id="default"),
        pytest.param({"observatory": GWOSC_GW150914_IFO[0]}, id="observatory"),
    ])
    def test_find(self, gw150914_16384, kwargs):
        """Test that `TimeSeries.find()` can actually find data."""
        ts = self.TEST_CLASS.find(
            GWOSC_GW150914_CHANNEL,
            *GWOSC_GW150914_SEGMENT,
            frametype=GWOSC_GW150914_FRAMETYPE,
            urltype="osdf",
            host="datafind.gwosc.org",
            **kwargs,
        )
        utils.assert_quantity_sub_equal(
            ts,
            gw150914_16384,
            exclude=["name", "channel", "unit"],
        )

    @_gwosc_pelican
    def test_find_datafind_httperror(self):
        """Test that HTTPErrors are presented in `find()`."""
        # Use an invalid observatory to trigger a 400 error from datafind
        with pytest.raises(ExceptionGroup, match="failed to get data") as excinfo:
            self.TEST_CLASS.find(
                GWOSC_GW150914_CHANNEL,
                *GWOSC_GW150914_SEGMENT,
                frametype=GWOSC_GW150914_FRAMETYPE,
                observatory="X",
                host="datafind.gwosc.org",
            )
        # Check that the 400 error is part of the exception group
        assert excinfo.group_contains(HTTPError, match="400 Client Error")

    @mock.patch.dict(
        "os.environ",
        {"GWDATAFIND_SERVER": GWOSC_DATAFIND_SERVER},
    )
    def test_find_datafind_runtimeerror(self):
        """Test that empty datafind caches result in RuntimeErrors in `find()`."""
        # Use a time where no data is available to trigger a RuntimeError
        with pytest.raises(ExceptionGroup, match="failed to get data") as excinfo:
            self.TEST_CLASS.find(
                GWOSC_GW150914_CHANNEL,
                *GWOSC_GW150914_SEGMENT.shift(-1e8),
                frametype=GWOSC_GW150914_FRAMETYPE,
                host="datafind.gwosc.org",
            )
        # Check that the RuntimeError is part of the exception group
        assert excinfo.group_contains(RuntimeError)

    def test_find_observatory_error(self):
        """Test that `find()` raises ValueError on inconsistent observatory."""
        with pytest.raises(ExceptionGroup, match="failed to get data") as excinfo:
            self.TEST_CLASS.find(
                "Test",
                0,
                1,
                frametype="X1_TEST",
            )
        assert excinfo.group_contains(
            ValueError,
            match="Cannot parse list of IFOs from channel names",
        )

    @_gwosc_pelican
    def test_find_best_frametype_in_find(self, gw150914_16384):
        """Test that `TimeSeries.find()` best frametype selection works."""
        ts = self.TEST_CLASS.find(
            GWOSC_GW150914_CHANNEL,
            GWOSC_GW150914_SEGMENT.start,
            GWOSC_GW150914_SEGMENT.end,
            urltype="osdf",
            host="datafind.gwosc.org",
        )
        utils.assert_quantity_sub_equal(
            ts,
            gw150914_16384,
            exclude=["name", "channel", "unit"],
        )

    @_gwosc_pelican
    def test_get_datafind(self, gw150914_16384):
        """Test that `TimeSeries.get(..., source='datafind')` works."""
        ts = self.TEST_CLASS.get(
            GWOSC_GW150914_CHANNEL,
            GWOSC_GW150914_SEGMENT.start,
            GWOSC_GW150914_SEGMENT.end,
            source="gwdatafind",
            host="datafind.gwosc.org",
            frametype_match=r"V1\Z",
            urltype="osdf",
        )
        utils.assert_quantity_sub_equal(
            ts,
            gw150914_16384,
            exclude=["name", "channel", "unit"],
        )

    @pytest_skip_flaky_network
    def test_get_gwosc_kwargs(self, gw150914):
        """Test that `TimeSeries.get(..., frametype="X")` doesn't break GWOSC.

        GWDataFind should be tried first as a source, but if the frametype doesn't
        match anything, it should fall back to the GWOSC API without falling over.
        """
        try:
            ts = self.TEST_CLASS.get(
                GWOSC_GW150914_IFO,
                GWOSC_GW150914_SEGMENT.start,
                GWOSC_GW150914_SEGMENT.end,
                frametype="WONT_MATCH",
                sample_rate=4096,
            )
        except* ImportError as e:  # pragma: no-cover
            pytest.skip(str(e))
        utils.assert_quantity_sub_equal(
            ts,
            gw150914,
            exclude=["name", "channel", "unit"],
        )

    @pytest_skip_flaky_network
    @pytest.mark.requires("nds2")
    @utils.skip_kerberos_credential
    def test_get_nds2(self, gw150914_16384):
        """Test that `TimeSeries.get(..., source="nds2")` works."""
        # get using NDS2 (if datafind could have been used to start with)
        with mock.patch("gwpy.timeseries.TimeSeries.fetch") as mock_fetch:
            ts = self.TEST_CLASS.get(
                NDS2_GW150914_CHANNEL,
                *GWOSC_GW150914_SEGMENT,
                source="nds2",
            )
        mock_fetch.assert_not_called()
        utils.assert_quantity_sub_equal(
            ts,
            gw150914_16384,
            exclude=["name", "channel", "unit"],
        )

    # -- signal processing methods ---

    def test_fft(self, gw150914):
        """Test `TimeSeries.fft`."""
        fs = gw150914.fft()
        assert isinstance(fs, FrequencySeries)
        assert fs.size == gw150914.size // 2 + 1
        assert fs.f0 == 0 * units.Hz
        assert fs.df == 1 / gw150914.duration
        assert fs.channel is gw150914.channel
        nptest.assert_almost_equal(
            fs.value.max(), 9.793003238789471e-20+3.5377863373683966e-21j)

        # test with nfft arg
        fs = gw150914.fft(nfft=256)
        assert fs.size == 129
        assert fs.dx == gw150914.sample_rate / 256

    @pytest.mark.parametrize("data", [
        [1., 0., -1., 0.],
        [1., 2., 3., 2., 1., 0.],
        numpy.arange(10),
        RNG.random(size=100),
    ])
    def test_fft_ifft(self, data):
        """Test that `TimeSeries.fft().ifft()` returns the original data."""
        a = self.TEST_CLASS(data)
        utils.assert_quantity_sub_equal(
            a,
            a.fft().ifft(),
            almost_equal=True,
            rtol=1e-7,
            atol=1e-10,
        )

    def test_average_fft(self, gw150914):
        """Test `TimeSeries.average_fft()`."""
        # test all defaults
        fs = gw150914.average_fft()
        utils.assert_quantity_sub_equal(fs, gw150914.detrend().fft())

        # test fftlength
        fs = gw150914.average_fft(fftlength=0.5)
        assert fs.size == 0.5 * gw150914.sample_rate.value // 2 + 1
        assert fs.df == 2 * units.Hertz

        fs = gw150914.average_fft(fftlength=0.4, overlap=0.2)

    def test_psd_default_overlap(self, gw150914):
        """Test that `TimeSeries.psd()` default overlap is half fftlength."""
        utils.assert_quantity_sub_equal(
            gw150914.psd(.5, method="median", window="hann"),
            gw150914.psd(.5, .25, method="median", window="hann"),
        )

    @pytest.mark.requires("lal")
    def test_psd_lal_median_mean(self, gw150914):
        """Test `TimeSeries.psd(method='lal_median_mean')`."""
        # check that warnings and errors get raised in the right place
        # for a median-mean PSD with the wrong data size or parameters

        # single segment should raise error
        with (
            pytest.raises(ValueError, match="Cannot calculate median-mean spectrum"),
            pytest.deprecated_call(),
        ):
            gw150914.psd(abs(gw150914.span), method="lal_median_mean")

        # odd number of segments should warn
        with (
            pytest.warns(UserWarning, match="Data array is the wrong size"),
            pytest.deprecated_call(),
        ):
            gw150914.psd(1, .5, method="lal_median_mean")

    @pytest.mark.parametrize("method", ["welch", "bartlett", "median"])
    def test_psd(self, noisy_sinusoid, method):
        """Test `TimeSeries.psd()`."""
        fftlength = .5
        overlap = .25
        fs = noisy_sinusoid.psd(
            fftlength=fftlength,
            overlap=overlap,
            method=method,
        )
        assert fs.unit == noisy_sinusoid.unit ** 2 / units.Hz
        assert fs.max() == fs.value_at(500)
        assert fs.size == fftlength * noisy_sinusoid.sample_rate.value // 2 + 1
        assert fs.f0 == 0 * units.Hz
        assert fs.df == units.Hz / fftlength
        assert fs.name == noisy_sinusoid.name
        assert fs.channel is noisy_sinusoid.channel

    @pytest.mark.parametrize(("library", "method"), chain(
        product(["pycbc.psd"], ["welch", "bartlett", "median", "median_mean"]),
        product(["lal"], ["welch", "bartlett", "median", "median_mean"]),
    ))
    def test_psd_deprecated(self, noisy_sinusoid, library, method):
        """Test deprecated average methods for TimeSeries.psd."""
        pytest.importorskip(library)

        fftlength = .5
        overlap = .25

        # remove final .25 seconds to stop median-mean complaining
        # (means an even number of overlapping FFT segments)
        if method == "median_mean":
            end = noisy_sinusoid.span[1]
            noisy_sinusoid = noisy_sinusoid.crop(end=end-overlap)

        # get actual method name
        library = library.split(".", 1)[0]

        with pytest.deprecated_call():
            psd = noisy_sinusoid.psd(
                fftlength=fftlength,
                overlap=overlap,
                method=f"{library}-{method}",
            )

        assert isinstance(psd, FrequencySeries)
        assert psd.unit == noisy_sinusoid.unit ** 2 / units.Hz
        assert psd.max() == psd.value_at(500)

    def test_asd(self, gw150914):
        """Test that `TimeSeries.asd()` is the sqrt of `psd()`."""
        kw = {
            "method": "median",
        }
        utils.assert_quantity_sub_equal(
            gw150914.asd(1, **kw),
            gw150914.psd(1, **kw) ** (1/2.),
        )

    def test_csd(self, noisy_sinusoid):
        """Test that `TimeSeries.csd(self)` is the same as `psd()`."""
        fs = noisy_sinusoid.csd(noisy_sinusoid, average="mean")
        utils.assert_quantity_sub_equal(
            fs.abs(),
            noisy_sinusoid.psd(method="welch"),
            exclude=["name"],
            almost_equal=True,
        )

    def test_csd_fftlength(self, noisy_sinusoid, corrupt_noisy_sinusoid):
        """Test that `TimeSeries.csd` uses the ``fftlength`` keyword properly."""
        # test fftlength is used
        fs = noisy_sinusoid.csd(corrupt_noisy_sinusoid, fftlength=0.5)
        assert fs.size == 0.5 * noisy_sinusoid.sample_rate.value // 2 + 1
        assert fs.df == 2 * units.Hertz

        # test that the default overlap works (by comparing to explicit)
        utils.assert_quantity_sub_equal(
            fs,
            noisy_sinusoid.csd(
                corrupt_noisy_sinusoid,
                fftlength=0.5,
                overlap=0.25,
            ),
        )

    @staticmethod
    def _window_helper(series, fftlength, window="hamming"):
        nfft = int(series.sample_rate.value * fftlength)
        return signal.get_window(window, nfft)

    @pytest.mark.parametrize("method", [
        "scipy-welch",
        "scipy-bartlett",
        "scipy-median",
        pytest.param("lal-welch", marks=pytest.mark.requires("lal")),
        pytest.param("lal-bartlett", marks=pytest.mark.requires("lal")),
        pytest.param("lal-median", marks=pytest.mark.requires("lal")),
        pytest.param("pycbc-welch", marks=pytest.mark.requires("pycbc.psd")),
        pytest.param(
            "pycbc-bartlett",
            marks=pytest.mark.requires("pycbc.psd"),
        ),
        pytest.param("pycbc-median", marks=pytest.mark.requires("pycbc.psd")),
    ])
    @pytest.mark.parametrize("window", [None, "hann", ("kaiser", 24), "array"])
    def test_spectrogram(self, gw150914, method, window):
        """Test `TimeSeries.spectrogram()`."""
        # generate window for 'array'
        win = self._window_helper(gw150914, 1) if window == "array" else window

        if method.startswith(("lal", "pycbc")):
            ctx = pytest.deprecated_call  # ignore: type[assignment]
        else:
            ctx = nullcontext  # ignore: type[assignment]

        # generate spectrogram
        with ctx():
            sg = gw150914.spectrogram(1, method=method, window=win)

        # validate
        assert isinstance(sg, Spectrogram)
        assert sg.shape == (
            abs(gw150914.span),
            gw150914.sample_rate.value // 2 + 1,
        )
        assert sg.f0 == 0 * units.Hz
        assert sg.df == 1 * units.Hz
        assert sg.channel is gw150914.channel
        assert sg.unit == gw150914.unit ** 2 / units.Hz
        assert sg.epoch == gw150914.epoch
        assert sg.span == gw150914.span

        # check the first time-bin is the same result as .psd()
        n = int(gw150914.sample_rate.value)
        if window == "hann" and not method.endswith("bartlett"):
            n = int(n * 1.5)  # default is 50% overlap
        with ctx():
            psd = gw150914[:n].psd(fftlength=1, method=method, window=win)
        # FIXME: epoch should not be excluded here (probably)
        utils.assert_quantity_sub_equal(
            sg[0],
            psd,
            exclude=["epoch"],
            almost_equal=True,
        )

    def test_spectrogram_fftlength(self, gw150914):
        """Test `TimeSeries.spectrogram()` uses the ``fftlength`` keyword properly."""
        sg = gw150914.spectrogram(1, fftlength=0.5, method="median")
        assert sg.shape == (
            abs(gw150914.span),
            0.5 * gw150914.sample_rate.value // 2 + 1,
        )
        assert sg.df == 2 * units.Hertz
        assert sg.dt == 1 * units.second

    def test_spectrogram_overlap(self, gw150914):
        """Test `TimeSeries.spectrogram()` uses the ``overlap`` keyword properly."""
        kw = {
            "fftlength": 0.5,
            "window": "hann",
            "method": "median",
        }
        sg = gw150914.spectrogram(1, **kw)
        sg2 = gw150914.spectrogram(1, overlap=.25, **kw)
        utils.assert_quantity_sub_equal(sg, sg2, almost_equal=True)

    def test_spectrogram_multiprocessing(self, gw150914):
        """Test `TimeSeries.spectrogram()` with multiprocessing."""
        kw = {
            "fftlength": 0.5,
            "window": "hann",
            "method": "median",
        }
        sg = gw150914.spectrogram(1, **kw)
        sg2 = gw150914.spectrogram(1, nproc=2, **kw)
        utils.assert_quantity_sub_equal(sg, sg2, almost_equal=True)

    @pytest.mark.parametrize("library", [
        pytest.param("lal", marks=pytest.mark.requires("lal")),
        pytest.param("pycbc", marks=pytest.mark.requires("pycbc.psd")),
    ])
    def test_spectrogram_median_mean(self, gw150914, library):
        """Test `TimeSeries.spectrogram()` with (deprecated) median-mean."""
        method = f"{library}-median-mean"

        # the LAL implementation of median-mean warns if not given the
        # correct amount of data for an even number of FFTs.
        if library == "lal":
            lal_warn_ctx = pytest.warns(
                UserWarning,
                match="Data array is the wrong size",
            )
        else:
            lal_warn_ctx = nullcontext()  # ignore: type[assignment]

        with pytest.deprecated_call(), lal_warn_ctx:
            sg = gw150914.spectrogram(
                1.5,
                fftlength=.5,
                overlap=0,
                method=method,
            )

        assert sg.dt == 1.5 * units.second
        assert sg.df == 2 * units.Hertz

    def test_spectrogram2(self, gw150914):
        """Test `TimeSeries.spectrogram2()`."""
        # test defaults
        sg = gw150914.spectrogram2(1, overlap=0)
        utils.assert_quantity_sub_equal(
            sg,
            gw150914.spectrogram(
                1,
                fftlength=1,
                overlap=0,
                method="scipy-welch",
                window="hann",
            ),
        )

        # test fftlength
        sg = gw150914.spectrogram2(0.5)
        assert sg.shape == (16, 0.5 * gw150914.sample_rate.value // 2 + 1)
        assert sg.df == 2 * units.Hertz
        assert sg.dt == 0.25 * units.second
        # test overlap
        sg = gw150914.spectrogram2(fftlength=0.25, overlap=0.24)
        assert sg.shape == (399, 0.25 * gw150914.sample_rate.value // 2 + 1)
        assert sg.df == 4 * units.Hertz
        # note: bizarre stride length because 4096/100 gets rounded
        assert sg.dt == 0.010009765625 * units.second

    def test_fftgram(self, gw150914):
        """Test `TimeSeries.fftgram()`."""
        fgram = gw150914.fftgram(1)
        fs = int(gw150914.sample_rate.value)
        f, t, sxx = signal.spectrogram(
            gw150914,
            fs,
            window="hann",
            nperseg=fs,
            mode="complex",
        )
        utils.assert_array_equal(gw150914.t0.value + t, fgram.xindex.value)
        utils.assert_array_equal(f, fgram.yindex.value)
        utils.assert_array_equal(sxx.T, fgram.value)

    def test_fftgram_overlap(self, gw150914):
        """Test `TimeSeries.fftgram()` with overlap."""
        fgram = gw150914.fftgram(1, overlap=0.5)
        fs = int(gw150914.sample_rate.value)
        f, t, sxx = signal.spectrogram(
            gw150914,
            fs,
            window="hann",
            nperseg=fs,
            noverlap=fs//2,
            mode="complex",
        )
        utils.assert_array_equal(gw150914.t0.value + t, fgram.xindex.value)
        utils.assert_array_equal(f, fgram.yindex.value)
        utils.assert_array_equal(sxx.T, fgram.value)

    def test_spectral_variance(self, gw150914):
        """Test `TimeSeries.spectral_variance()`."""
        variance = gw150914.spectral_variance(.5, method="median")
        assert isinstance(variance, SpectralVariance)
        assert variance.x0 == 0 * units.Hz
        assert variance.dx == 2 * units.Hz
        assert variance.max() == 8

    def test_rayleigh_spectrum(self, gw150914):
        """Test `TimeSeries.rayleigh_spectrum()`."""
        # assert single FFT creates Rayleigh of 0
        ray = gw150914.rayleigh_spectrum()
        assert isinstance(ray, FrequencySeries)
        assert ray.unit is units.Unit("")
        assert ray.name == f"Rayleigh spectrum of {gw150914.name}"
        assert ray.epoch == gw150914.epoch
        assert ray.channel is gw150914.channel
        assert ray.f0 == 0 * units.Hz
        assert ray.df == 1 / gw150914.duration
        assert ray.sum().value == 0

        # actually test properly
        ray = gw150914.rayleigh_spectrum(.5)  # no overlap
        assert ray.df == 2 * units.Hz
        nptest.assert_almost_equal(ray.max().value, 2.1239253590490157)
        assert ray.frequencies[ray.argmax()] == 1322 * units.Hz

        ray = gw150914.rayleigh_spectrum(.5, .25)  # 50 % overlap
        nptest.assert_almost_equal(ray.max().value, 1.8814775174483833)
        assert ray.frequencies[ray.argmax()] == 136 * units.Hz

    def test_csd_spectrogram(self, gw150914):
        """Test `TimeSeries.csd_spectrogram()`."""
        # test defaults
        sg = gw150914.csd_spectrogram(gw150914, 1)
        assert isinstance(sg, Spectrogram)
        assert sg.shape == (4, gw150914.sample_rate.value // 2 + 1)
        assert sg.f0 == 0 * units.Hz
        assert sg.df == 1 * units.Hz
        assert sg.channel is gw150914.channel
        assert sg.unit == gw150914.unit ** 2 / units.Hertz
        assert sg.epoch == gw150914.epoch
        assert sg.span == gw150914.span

        # check the same result as CSD
        cropped = gw150914[:int(gw150914.sample_rate.value)]
        csd = cropped.csd(cropped)
        utils.assert_quantity_sub_equal(sg[0], csd, exclude=["name", "epoch"])

        # test fftlength
        sg = gw150914.csd_spectrogram(gw150914, 1, fftlength=0.5)
        assert sg.shape == (4, 0.5 * gw150914.sample_rate.value // 2 + 1)
        assert sg.df == 2 * units.Hertz
        assert sg.dt == 1 * units.second

        # test overlap
        sg = gw150914.csd_spectrogram(
            gw150914,
            0.5,
            fftlength=0.25,
            overlap=0.125,
        )
        assert sg.shape == (8, 0.25 * gw150914.sample_rate.value // 2 + 1)
        assert sg.df == 4 * units.Hertz
        assert sg.dt == 0.5 * units.second

        # test multiprocessing
        sg2 = gw150914.csd_spectrogram(
            gw150914,
            0.5,
            fftlength=0.25,
            overlap=0.125,
            nproc=2,
        )
        utils.assert_quantity_sub_equal(sg, sg2)

    @pytest.mark.parametrize("ftype", ["fir", "iir"])
    @pytest.mark.parametrize("rate", [
        1024,
        pytest.param(units.Quantity(1024, "Hz"), id="1024 Hz"),
        pytest.param(units.Quantity(1.5, "kHz"), id="1.5 kHz"),
    ])
    def test_resample(self, gw150914, rate, ftype):
        """Test `gwpy.timeseries.TimeSeries.resample`."""
        # resample
        l2 = gw150914.resample(rate, ftype=ftype)
        assert l2.sample_rate == units.Quantity(rate, units.Hz)

        # compare ASDs
        asd1 = gw150914.asd(fftlength=4, overlap=2).crop(30, 400)
        asd2 = l2.asd(fftlength=4, overlap=2).crop(30, 400)
        max1 = asd1.argmax()
        max2 = asd2.argmax()
        # check that the peak occurs at the same frequency
        assert max1 == max2
        # check that the peak itself is approximately the same
        assert asd1[max1].value == pytest.approx(asd2[max2].value)

    def test_resample_simple_upsample(self, gw150914):
        """Test consistency when upsampling by 2x`."""
        upsamp = gw150914.resample(gw150914.sample_rate.value * 2)
        assert numpy.allclose(gw150914.value, upsamp.value[::2])

    def test_resample_noop(self):
        """Test that resampling to the same rate returns self with a warning."""
        data = self.TEST_CLASS([1, 2, 3, 4, 5])
        with pytest.warns(
            UserWarning,
            match=r"resample\(\) rate matches current sample_rate",
        ):
            new = data.resample(data.sample_rate)
        assert data is new

    def test_resample_rate_units(self):
        """Check that TimeSeries.resample fails on weird units."""
        with pytest.raises(
            ValueError,
            match="invalid resampling rate",
        ):
            self.TEST_CLASS([1, 2, 3, 4, 5]).resample(units.Quantity(1, "m"))

    def test_rms(self, gw150914):
        """Test `TimeSeries.rms()`."""
        rms = gw150914.rms(1.)
        assert rms.sample_rate == 1 * units.Hz

    @mock.patch("gwpy.segments.DataQualityFlag.query", return_value=LIVETIME)
    def test_mask(self, mock_query):
        """Test `TimeSeries.mask()`."""
        # craft a timeseries of ones that can be easily tested against
        # a few interesting corner cases
        data = TimeSeries(numpy.ones(8192), sample_rate=128)
        masked = data.mask(flag="X1:TEST-FLAG:1")

        # create tukey window for comparison
        window = tukey(128, alpha=1.0)
        times = (data.t0 + numpy.arange(data.size) * data.dt).value
        (live, ) = numpy.nonzero([t in LIVETIME.active for t in times])
        (dead, ) = numpy.nonzero([t not in LIVETIME.active for t in times])

        # verify the mask is correct
        assert data.is_compatible(masked)
        assert live.size + dead.size == data.size
        assert numpy.all(numpy.isfinite(masked.value[live]))
        assert numpy.all(numpy.isnan(masked.value[dead]))
        utils.assert_allclose(masked.value[:4032], numpy.ones(4032))
        # First segment [0, 32): only right side tapered, use last 64 samples
        utils.assert_allclose(masked.value[4032:4096], window[-64:])
        # Second segment [34, 34.5): both sides tapered
        utils.assert_allclose(
            masked.value[4352:4416],
            window[:64] * window[-64:],
        )

    def test_demodulate(self):
        """Test `TimeSeries.demodulate()`."""
        # create a timeseries that is simply one loud sinusoidal oscillation
        # at a particular frequency, then demodulate at that frequency and
        # recover the amplitude and phase
        amp, phase, f = 1., numpy.pi/4, 30
        duration, sample_rate, stride = 600, 4096, 60
        t = numpy.linspace(0, duration, duration*sample_rate)
        data = TimeSeries(amp * numpy.cos(2*numpy.pi*f*t + phase),
                          unit="", times=t)

        # test with exp=True
        demod = data.demodulate(f, stride=stride, exp=True)
        assert demod.unit == data.unit
        assert demod.size == duration // stride
        utils.assert_allclose(numpy.abs(demod.value), amp, rtol=1e-5)
        utils.assert_allclose(numpy.angle(demod.value), phase, rtol=1e-5)

        # test with exp=False, deg=True
        mag, ph = data.demodulate(f, stride=stride)
        assert mag.unit == data.unit
        assert mag.size == ph.size
        assert ph.unit == "deg"
        utils.assert_allclose(mag.value, amp, rtol=1e-5)
        utils.assert_allclose(ph.value, numpy.rad2deg(phase), rtol=1e-5)

        # test with exp=False, deg=False
        mag, ph = data.demodulate(f, stride=stride, deg=False)
        assert ph.unit == "rad"
        utils.assert_allclose(ph.value, phase, rtol=1e-5)

    def test_heterodyne(self):
        """Test `TimeSeries.heterodyne()`."""
        # create a timeseries that is simply one loud sinusoidal oscillation,
        # with a frequency and frequency derivative, then heterodyne using the
        # phase evolution to recover the amplitude and phase
        amp, phase, f, fdot = 1., numpy.pi/4, 30, 1e-4
        duration, sample_rate, stride = 600, 4096, 60
        t = numpy.linspace(0, duration, duration*sample_rate)
        phases = 2 * numpy.pi * (f * t + 0.5 * fdot * t ** 2)
        data = TimeSeries(
            amp * numpy.cos(phases + phase),
            unit="",
            times=t,
        )

        # test with default settings
        het = data.heterodyne(phases, stride=stride)
        assert het.unit == data.unit
        assert het.size == duration // stride
        utils.assert_allclose(numpy.abs(het.value), 0.5*amp, rtol=1e-4)
        utils.assert_allclose(numpy.angle(het.value), phase, rtol=2e-4)

        # test with singlesided=True
        het = data.heterodyne(
            phases,
            stride=stride,
            singlesided=True,
        )
        assert het.unit == data.unit
        assert het.size == duration // stride
        utils.assert_allclose(numpy.abs(het.value), amp, rtol=1e-4)
        utils.assert_allclose(numpy.angle(het.value), phase, rtol=2e-4)

    def test_heterodyne_scalar_phase_error(self, array):
        """Test `TimeSeries.heterodyne()` scalar phase error."""
        with pytest.raises(
            ValueError,
            match="Phase is not array_like",
        ):
            array.heterodyne(1.0)

    def test_heterodyne_phase_length_error(self, array):
        """Test `TimeSeries.heterodyne()` handling of a bad phase array."""
        with pytest.raises(
            ValueError,
            match="Phase array must be the same length",
        ):
            array.heterodyne(array[0:len(array) // 2])

    def test_taper(self):
        """Test `TimeSeries.taper()`."""
        # create a flat timeseries, then taper it
        t = numpy.linspace(0, 1, 2048)
        data = TimeSeries(numpy.cos(10*numpy.pi*t), times=t, unit="")
        tapered = data.taper()

        # check that the tapered timeseries goes to zero at its ends,
        # and that the operation does not change the original data
        assert tapered[0].value == 0
        assert tapered[-1].value == 0
        assert tapered.unit == data.unit
        assert tapered.size == data.size
        utils.assert_allclose(data.value, numpy.cos(10*numpy.pi*t))

        # run the same tests for a user-specified taper duration
        dtapered = data.taper(duration=0.1)
        assert dtapered[0].value == 0
        assert dtapered[-1].value == 0
        assert dtapered.unit == data.unit
        assert dtapered.size == data.size
        utils.assert_allclose(data.value, numpy.cos(10*numpy.pi*t))

        # run the same tests for a user-specified number of samples to taper
        stapered = data.taper(nsamples=10)
        assert stapered[0].value == 0
        assert stapered[-1].value == 0
        assert stapered.unit == data.unit
        assert stapered.size == data.size
        utils.assert_allclose(data.value, numpy.cos(10*numpy.pi*t))

    def test_gate(self):
        # generate Gaussian noise with std = 0.5
        noise = self.TEST_CLASS(
            numpy.random.normal(scale=0.5, size=16384*64),
            sample_rate=16384,
            epoch=-32,
        )
        # generate a glitch with amplitude 20 at 1000 Hz
        glitchtime = 0.0
        glitch = signal.gausspulse(
            noise.times.value - glitchtime,
            bw=100,
        ) * 20
        data = noise + glitch

        # check that the glitch is at glitchtime as expected
        tmax = data.times.value[data.argmax()]
        nptest.assert_almost_equal(tmax, glitchtime)

        # gating method will be called with whiten = False to decouple
        # whitening method from gating method
        tzero = 1.0
        tpad = 1.0
        threshold = 10.0
        gated = data.gate(tzero=tzero, tpad=tpad, threshold=threshold,
                          whiten=False)

        # check that the maximum value is not within the region set to zero
        tleft = glitchtime - tzero
        tright = glitchtime + tzero
        assert not tleft < gated.times.value[gated.argmax()] < tright

        # check that there are no remaining values above the threshold
        assert gated.max() < threshold

    def test_whiten(self, colored_noise, gausspulse):
        """Test `TimeSeries.whiten()`."""
        # Inject the pulse into the middle of the noise
        glitchtime = (colored_noise.t0 + colored_noise.duration / 2).value
        gausspulse.shift(glitchtime)
        data = colored_noise.inject(gausspulse * 1e-2)

        # When the input is stationary Gaussian noise, the output should have
        # zero mean and unit variance
        whitened = colored_noise.whiten(detrend="linear", method="median")
        assert whitened.size == colored_noise.size
        nptest.assert_almost_equal(whitened.mean().value, 0.0, decimal=1)
        nptest.assert_almost_equal(whitened.std().value, 1.0, decimal=1)
        tmax = data.times[data.argmax()].value
        assert tmax != glitchtime

        # When a loud signal is present,
        # the max amplitude should be recovered at the time of that signal
        whitened = data.whiten(detrend="linear", method="median")
        tmax = whitened.times[whitened.argmax()].value
        assert tmax == pytest.approx(glitchtime)

    def test_convolve(self):
        data = self.TEST_CLASS(
            signal.get_window("hann", 1024),
            sample_rate=512,
            epoch=-1,
        )
        filt = numpy.array([1, 0])

        # check that the 'valid' data are unchanged by this filter
        convolved = data.convolve(filt)
        assert convolved.size == data.size
        utils.assert_allclose(convolved.value[1:-1], data.value[1:-1])

    def test_correlate(self, colored_noise, gausspulse):
        # Inject the pulse into the middle of the noise
        glitchtime = (colored_noise.t0 + colored_noise.duration / 2).value
        gausspulse.shift(glitchtime)
        data = colored_noise.inject(gausspulse * 1e-2)

        # check that, without a signal present, we only see background
        snr = colored_noise.correlate(gausspulse, whiten=True, method="median")
        tmax = snr.times[snr.argmax()]
        assert snr.size == colored_noise.size
        assert not numpy.isclose(tmax.value, glitchtime)
        nptest.assert_almost_equal(snr.mean().value, 0.0, decimal=1)
        nptest.assert_almost_equal(snr.std().value, 1.0, decimal=1)

        # Recover the glitch
        snr = data.correlate(gausspulse, whiten=True, method="median")
        tmax = snr.times[snr.argmax()]
        nptest.assert_almost_equal(tmax.value, glitchtime)

    def test_detrend(self, gw150914):
        assert not numpy.isclose(gw150914.value.mean(), 0.0, atol=1e-21)
        detrended = gw150914.detrend()
        assert numpy.isclose(detrended.value.mean(), 0.0)

    def test_filter(self, gw150914):
        """Test `TimeSeries.filter()`."""
        fts = gw150914.filter(([], [], 1), analog=True)
        utils.assert_quantity_sub_equal(gw150914, fts)

    def test_filter_sos(self, gw150914):
        """Test `TimeSeries.filter()` with SOS filters."""
        zpk = filter_design.highpass(50, sample_rate=gw150914.sample_rate)
        sos = signal.zpk2sos(*zpk)
        utils.assert_quantity_almost_equal(
            gw150914.filter(zpk),
            gw150914.filter(sos),
        )

    def test_zpk(self, gw150914):
        """Test `TimeSeries.zpk()`.

        The zpk method is just a wrapper around filter, so we just
        check that they give the same result.
        """
        zpk = [10, 10], [1, 1], 100
        utils.assert_quantity_sub_equal(
            gw150914.zpk(*zpk, analog=True, unit="Hz"),
            gw150914.filter(zpk, analog=True, unit="Hz"),
        )

    def test_highpass_happy_path(self, gw150914):
        """Check that passband val are approx equal, stopband are not."""
        asd = gw150914.asd()
        hp_asd = gw150914.highpass(100).asd()

        eqinds = numpy.where(hp_asd.frequencies.value > 200)[0]
        eqind0 = eqinds[0]

        # be within 50% for all values after
        # (tolerance increased from 40% due to fix for numerical
        # stability in ZPK filter gain application - see issue #1544)
        # numpy allclose formula:
        # |a-b| <= atol + rtol * |b|

        assert numpy.allclose(
            hp_asd[eqind0:].value,
            asd[eqind0:].value,
            rtol=0.5,
            atol=0,
        )

        # dont be within 40% for all value before
        assert not numpy.allclose(
            hp_asd[:eqind0].value,
            asd[:eqind0].value,
            rtol=0.4,
            atol=0,
        )

    def test_lowpass_happy_path(self, gw150914):
        """Check that passband val are approx equal, stopband are not."""
        asd = gw150914.asd()
        lp_asd = gw150914.lowpass(500).asd()

        eqinds = numpy.where(lp_asd.frequencies.value < 500)[0]
        eqind0 = eqinds[0]

        # be within 40% for all values before
        # numpy allclose formula:
        # |a-b| <= atol + rtol * |b|

        assert not numpy.allclose(
            lp_asd[eqind0:].value,
            asd[eqind0:].value,
            rtol=0.4,
            atol=0,
        )

        # dont be within 40% for all value after
        assert numpy.allclose(
            lp_asd[:eqind0].value,
            asd[:eqind0].value,
            rtol=0.4,
            atol=0,
        )

    def test_notch_happy_path(self, gw150914):
        """Check passband vals are approx equal, stopband are not."""
        nf = 10
        notched = gw150914.notch(nf, filtfilt=True)
        notched_asd = notched.asd()
        asd = gw150914.asd()

        n_eps = 3
        eps = n_eps * notched_asd.df.value

        # indices outside interval around 10 rad/s
        l_inds = numpy.where(notched_asd.frequencies.value < nf - eps)[0]
        r_inds = numpy.where(notched_asd.frequencies.value > nf + eps)[0]
        # index at 10 rad/s
        nf_ind = numpy.argmin(numpy.abs(notched_asd.frequencies.value - nf))
        # indices inside interval around 10 rad/s
        nf_inds = numpy.arange(nf_ind - n_eps, nf_ind + n_eps)

        assert l_inds[-1] <= nf_ind
        assert r_inds[0] >= nf_ind

        # be within 40% for all values outside nbrhood
        assert numpy.allclose(
            notched_asd[l_inds].value,
            asd[l_inds].value,
            rtol=0.4,
            atol=0,
        )
        assert numpy.allclose(
            notched_asd[r_inds].value,
            asd[r_inds].value,
            rtol=0.4,
            atol=0,
        )

        # dont be within 40% for all values inside nbrhood
        assert not numpy.allclose(
            notched_asd[nf_inds].value,
            asd[nf_inds].value,
            rtol=0.4,
            atol=0,
        )

        # biggest difference between filtered and unfiltered
        # should be at closest f to nf=10
        absd = numpy.abs(notched_asd.value - asd.value)
        assert numpy.isclose(absd[nf_ind], numpy.max(absd))

    def test_bandpass_happy_path(self, gw150914):
        """Check that passband val are approx equal, stopband are not."""
        asd = gw150914.asd()
        bp_asd = gw150914.bandpass(100, 1000).asd()

        eqinds = numpy.where(
            numpy.logical_and(
                bp_asd.frequencies.value > 100,
                bp_asd.frequencies.value < 1000,
            ),
        )[0]

        eqind0 = eqinds[0]
        eqindn = eqinds[-1]

        # be within 50% for all values in passband
        # (tolerance increased from 40% due to fix for numerical
        # stability in ZPK filter gain application - see issue #1544)
        # numpy allclose formula:
        # |a-b| <= atol + rtol * |b|

        assert numpy.allclose(
            bp_asd[eqind0:eqindn].value,
            asd[eqind0:eqindn].value,
            rtol=0.5,
            atol=0,
        )

        # dont be within 40% for all value before
        assert not numpy.allclose(
            bp_asd[:eqind0].value,
            asd[:eqind0].value,
            rtol=0.4,
            atol=0,
        )

        # or after
        assert not numpy.allclose(
            bp_asd[eqindn:].value,
            asd[eqindn:].value,
            rtol=0.4,
            atol=0,
        )

    def test_notch(self, gw150914):
        """Test `TimeSeries.notch()`."""
        # test notch runs end-to-end
        gw150914.notch(60)

        # test breaks when you try and 'fir' notch
        with pytest.raises(NotImplementedError):
            gw150914.notch(10, type="fir")

    def test_q_gram(self, gw150914):
        """Test `TimeSeries.q_gram()`."""
        qgram = gw150914.q_gram()
        assert isinstance(qgram, EventTable)
        assert qgram.meta["q"] == 45.25483399593904
        assert qgram["energy"].min() >= 5.5**2 / 2
        nptest.assert_almost_equal(qgram["energy"].max(), 10559.25, decimal=2)

    def test_q_transform(self, gw150914):
        """Test `TimeSeries.q_transform()`."""
        qspecgram = gw150914.q_transform(method="scipy-welch", fftlength=2)
        assert isinstance(qspecgram, Spectrogram)
        assert qspecgram.shape == (1000, 2403)
        assert qspecgram.q == 5.65685424949238
        nptest.assert_almost_equal(qspecgram.value.max(), 155.93567, decimal=5)
        nptest.assert_almost_equal(
            qspecgram.value.mean(),
            1.936469,
            decimal=5,
        )

        # test whitening args
        asd = gw150914.asd(2, 1, method="scipy-welch")
        qsg2 = gw150914.q_transform(method="scipy-welch", whiten=asd)
        utils.assert_quantity_sub_equal(qspecgram, qsg2, almost_equal=True)

        asd = gw150914.asd(.5, .25, method="scipy-welch")
        qsg2 = gw150914.q_transform(method="scipy-welch", whiten=asd)
        qsg3 = gw150914.q_transform(
            method="scipy-welch",
            fftlength=.5,
            overlap=.25,
        )
        utils.assert_quantity_sub_equal(qsg2, qsg3, almost_equal=True)

        # make sure frequency too high presents warning
        with pytest.warns(
            UserWarning,
            match="upper frequency of 10000.0 Hz is too high",
        ):
            qspecgram = gw150914.q_transform(
                method="scipy-welch",
                frange=(0, 10000),
            )
        nptest.assert_almost_equal(
            qspecgram.yspan[1],
            1291.5316,
            decimal=4,
        )

        # test other normalisations work (or don't)
        q2 = gw150914.q_transform(method="scipy-welch", norm="median")
        utils.assert_quantity_sub_equal(qspecgram, q2, almost_equal=True)
        gw150914.q_transform(method="scipy-welch", norm="mean")
        gw150914.q_transform(method="scipy-welch", norm=False)
        with pytest.raises(
            ValueError,
            match=r"^invalid normalisation 'blah'$",
        ):
            gw150914.q_transform(method="scipy-welch", norm="blah")

    def test_q_transform_logf(self, gw150914):
        """Test `TimeSeries.q_transform()` with log frequency spacing."""
        qspecgram = gw150914.q_transform(
            method="scipy-welch",
            fftlength=2,
            logf=True,
        )
        assert isinstance(qspecgram, Spectrogram)
        assert qspecgram.shape == (1000, 500)
        assert qspecgram.q == 5.65685424949238
        nptest.assert_almost_equal(qspecgram.value.max(), 155.93774, decimal=5)

    def test_q_transform_nan(self):
        """Test that `TimeSeries.q_transform()` raises for NaN input."""
        data = TimeSeries(numpy.empty(256*10) * numpy.nan, sample_rate=256)
        with pytest.raises(
            ValueError,
            match=r"^Input signal contains non-numerical values$",
        ):
            data.q_transform(method="median")

    def test_boolean_statetimeseries(self, array):
        """Test comparisons of `TimeSeries` returning `StateTimeSeries`."""
        comp = array >= 2 * array.unit
        assert isinstance(comp, StateTimeSeries)
        assert comp.unit is units.Unit("")
        assert comp.name == f"{array.name} >= 2.0"
        assert (array == array).name == f"{array.name} == {array.name}"  # noqa: PLR0124

    @pytest_skip_flaky_network
    def test_transfer_function(self, gw150914_h1_32, gw150914_l1_32):
        """Test `TimeSeries.transfer_function()`."""
        tf = gw150914_h1_32.transfer_function(
            gw150914_l1_32,
            fftlength=1.0,
            overlap=0.5,
        )
        assert tf.df == 1 * units.Hz
        assert tf.frequencies[abs(tf).argmax()] == 516 * units.Hz

    @pytest_skip_flaky_network
    def test_coherence(self, gw150914_h1_32, gw150914_l1_32):
        """Test `TimeSeries.coherence()`."""
        coh = gw150914_h1_32.coherence(
            gw150914_l1_32,
            fftlength=1.0,
        )
        assert coh.df == 1 * units.Hz
        assert coh.frequencies[coh.argmax()] == 60 * units.Hz

    @pytest_skip_flaky_network
    def test_coherence_spectrogram(self, gw150914_h1_32, gw150914_l1_32):
        """Test `TimeSeries.coherence_spectrogram()`."""
        cohsg = gw150914_h1_32.coherence_spectrogram(
            gw150914_l1_32,
            4,
            fftlength=1.0,
        )
        assert cohsg.t0 == gw150914_h1_32.t0
        assert cohsg.dt == 4 * units.second
        assert cohsg.df == 1 * units.Hz
        _, fmax = numpy.unravel_index(cohsg.argmax(), cohsg.shape)
        assert cohsg.frequencies[fmax] == 60 * units.Hz


# -- TimeSeriesDict ------------------

class TestTimeSeriesDict(_TestTimeSeriesBaseDict[TimeSeriesDict, TimeSeries]):
    """Tests for `gwpy.timeseries.TimeSeriesDict`."""

    channels: ClassVar[list[str]] = ["H1:LDAS-STRAIN", "L1:LDAS-STRAIN"]
    TEST_CLASS = TimeSeriesDict
    ENTRY_CLASS = TimeSeries

    @pytest.mark.requires("framel")
    def test_read_write_gwf(self, instance, tmp_path):
        """Test reading and writing GWF files."""
        tmp = tmp_path / "test.gwf"
        instance.write(tmp)
        new = self.TEST_CLASS.read(tmp, instance.keys())
        for key in new:
            utils.assert_quantity_sub_equal(new[key], instance[key],
                                            exclude=["channel"])

    def test_read_write_hdf5(self, instance, tmp_path):
        """Test reading and writing HDF5 files."""
        tmp = tmp_path / "test.h5"
        instance.write(tmp, overwrite=True)
        new = self.TEST_CLASS.read(tmp, instance.keys())
        for key in new:
            utils.assert_quantity_sub_equal(new[key], instance[key])
        # check auto-detection of names
        new = self.TEST_CLASS.read(tmp)
        for key in new:
            utils.assert_quantity_sub_equal(new[key], instance[key])


# -- TimeSeriesList ------------------

class TestTimeSeriesList(_TestTimeSeriesBaseList[TimeSeriesList, TimeSeries]):
    """Tests for :class:`gwpy.timeseries.TimeSeriesList`."""

    TEST_CLASS = TimeSeriesList
    ENTRY_CLASS = TimeSeries



# Copyright (c) 2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave frame (GWF) files using the FrameL API.

The frame format is defined in :dcc:`LIGO-T970130`.
"""

from __future__ import annotations

import warnings
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    TypedDict,
)

import framel
from igwn_segments import infinity

from ....io.gwf.core import _series_name
from ....io.utils import (
    file_list,
    file_path,
)
from ....segments import Segment
from ... import (
    TimeSeries,
    TimeSeriesBaseList,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path
    from typing import (
        IO,
        TypeVar,
    )

    import numpy

    from ....time import LIGOTimeGPS
    from ... import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

    _TimeSeriesType = TypeVar("_TimeSeriesType", bound=TimeSeriesBase)

inf = infinity()

FRAMEL_COMPRESSION_GZIP = {
    "gzip",
    1,
    257,
    None,
}

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

def read(
    source: str | Path | IO,
    channels: list[str],
    start: LIGOTimeGPS | None = None,
    end: LIGOTimeGPS | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> TimeSeriesBaseDict:
    """Read data from one or more GWF files using the FrameL API."""
    # scaled must be provided to provide a consistent API with frameCPP
    if scaled is not None:
        warnings.warn(
            "the `scaled` keyword argument is not supported by framel, "
            "if you require ADC scaling, please install "
            "python-ldas-tools-framecpp",
            stacklevel=2,
        )

    # parse input source
    files = file_list(source)

    # read each file and channel individually and append
    tmp: dict[str, TimeSeriesBaseList] = defaultdict(TimeSeriesBaseList)
    for file_ in files:
        for series in _read_gwf(
            file_,
            channels,
            start,
            end,
            series_class,
        ):
            tmp[series.name].append(series)
    out = series_class.DictClass()
    for channel, serieslist in tmp.items():
        out[channel] = serieslist.join()

    return out


def _read_gwf(
    filename: str,
    channels: list[str],
    start: LIGOTimeGPS | None,
    end: LIGOTimeGPS | None,
    series_class: type[TimeSeriesBase],
) -> Iterator[TimeSeriesBase]:
    """Read data from a single file."""
    if start is None:
        start = -inf
    if end is None:
        end = inf
    span = Segment(start, end)
    record: set[str] = set()
    _record = record.add
    for name in channels:
        # framel.frgetvect will silently pad available data to
        # whatever span you gave it without error, so we have to
        # read everything and crop
        new = _read_channel(
            filename,
            name,
            -1,  # read from start
            -1,  # read all data
            series_class,
        )
        try:
            overlap = new.span & span
        except ValueError:
            # this channel doesn't overlap with the requested span
            continue
        if (keep := new.crop(*overlap)).size:
            yield keep
            _record(keep.name)

    # if any channels weren't read, something went wrong
    for channel in channels:
        if (name := str(channel)) not in record:
            msg = f"cannot read data for '{name}' from '{filename}' in interval {span}"
            raise ValueError(msg)


def _read_channel(
    filename: str | Path,
    channel: str,
    start: float,
    duration: float,
    series_class: type[_TimeSeriesType],
) -> _TimeSeriesType:
    """Read one channel from one file."""
    try:
        data, gps, offset, dx, xunit, yunit = framel.frgetvect1d(
            str(filename),
            str(channel),
            start=start,
            span=duration,
        )
    except KeyError as exc:  # upstream errors
        raise ValueError(str(exc)) from exc
    except ValueError as exc:
        if str(exc) == "NULL pointer access":
            msg = f"channel '{channel}' not found"
            raise ValueError(msg) from exc
        raise
    return series_class(
        data,
        name=channel,
        x0=gps+offset,
        dx=dx,
        xunit=xunit,
        unit=yunit,
    )


# -- write ---------------------------

def write(
    tsdict: TimeSeriesBaseDict,
    outfile: str | Path | IO,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    type: str | None = None,
    name: str | None = None,
    run: int = 0,
    compression: int | str | None = None,
    compression_level: int | None = None,
) -> None:
    """Write data to a GWF file using the FrameL API."""
    if name is not None:
        warnings.warn(
            "python-framel does not support setting FrHistory 'name', "
            "this value will be ignored",
            stacklevel=2,
        )
    if run:
        warnings.warn(
            "python-framel does not support setting FrHistory 'run', "
            "this value will be ignored",
            stacklevel=2,
        )

    # FrameL Python bindings only support Gzip
    if isinstance(compression, str):
        compression = compression.lower()
    if compression not in FRAMEL_COMPRESSION_GZIP:
        msg = "python-framel only supports compression='GZIP'"
        raise ValueError(msg)
    if compression_level not in (None, 1):
        warnings.warn(
            "python-framel only supports GZIP level 1",
            stacklevel=1,
        )

    # format and crop each series
    channellist: list[FrameLVectDict] = [
        _channel_data_to_write(
            series.crop(start=start, end=end),
            type,
        ) for series in tsdict.values()
    ]
    return framel.frputvect(file_path(outfile), channellist)


class FrameLVectDict(TypedDict):
    """Dict of channel vector information."""

    name: str
    data: numpy.ndarray
    start: float
    dx: float
    x_unit: str
    y_unit: str
    kind: str
    type: int
    subType: int


def _channel_data_to_write(
    timeseries: TimeSeriesBase,
    type_: str | None,
) -> FrameLVectDict:
    """Format a series into a dict suitable for `framel.frputvect`."""
    return FrameLVectDict({
        "name": _series_name(timeseries) or "",
        "data": timeseries.value,
        "start": timeseries.x0.to("s").value,
        "dx": timeseries.dx.to("s").value,
        "x_unit": str(timeseries.xunit),
        "y_unit": str(timeseries.unit),
        "kind": (
            type_
            or getattr(timeseries.channel, "_ctype", "proc")
            or "proc"
        ).upper(),
        "type": 1,
        "subType": 0,
    })



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Spectrogram plots."""

from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    cast,
)

from numpy import percentile

from .cliproduct import (
    FFTMixin,
    ImageProduct,
    TimeDomainProduct,
)

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from typing import ClassVar

    from astropy.units import UnitBase
    from matplotlib.collections import Collection
    from matplotlib.image import AxesImage

    from ..plot import (
        Axes,
        Plot,
    )
    from ..spectrogram import Spectrogram

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = logging.getLogger(__name__)

#: Maximum number of FFTs to calculate for high-resolution spectrograms
MAX_FFT_PER_SECOND: int = 3


class SpectrogramProduct(FFTMixin, TimeDomainProduct, ImageProduct):
    """Plot the spectrogram of a time series."""

    action: ClassVar[str] = "spectrogram"

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Initialise this `SpectrogramProduct`."""
        super().__init__(args, logger=logger)

        #: attribute to hold calculated Spectrogram data array
        self.result: Spectrogram | None = None

    @classmethod
    def arg_yaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Configure Y-axis arguments."""
        return cls._arg_faxis("y", parser)

    def _finalize_arguments(self, args: Namespace) -> None:
        """Sanity-check and set defaults for arguments."""
        if args.color_scale is None:
            args.color_scale = "log"
        super()._finalize_arguments(args)

    @property
    def units(self) -> list[UnitBase]:
        """The (unique) list of data units for this product."""
        if self.result is None:
            return [None]
        return [self.result.unit]

    def get_ylabel(self) -> str | None:
        """Return the label for the Y-axis."""
        return "Frequency (Hz)"

    def get_title(self) -> str:
        """Return default title for plot."""
        return f"fftlength={self.args.secpfft}, overlap={self.args.overlap}"

    def get_suptitle(self) -> str:
        """Return default super-title for plot."""
        return f"Spectrogram: {self.chan_list[0]}"

    def get_color_label(self) -> str | None:
        """Text for colorbar label."""
        if self.args.norm:
            return f"Normalized to {self.args.norm}"
        if len(self.units) == 1 and self.usetex:
            u = self.units[0].to_string("latex").strip("$")
            return fr"ASD $\left({u}\right)$"
        if len(self.units) == 1:
            u = self.units[0].to_string("generic")
            return f"ASD ({u})"
        return super().get_color_label()

    def get_stride(self) -> float | None:
        """Calculate the stride for the spectrogram.

        This method returns the stride as a `float`, or `None` to indicate
        selected usage of `TimeSeries.spectrogram2`.
        """
        fftlength = float(self.args.secpfft)
        overlap = fftlength * self.args.overlap
        stride = fftlength - overlap
        nfft = self.duration / stride  # number of FFTs
        ffps = int(nfft / (self.width * 0.8))  # FFTs per second
        if ffps > MAX_FFT_PER_SECOND:
            return max(2 * fftlength, ffps * stride + fftlength - 1)
        return None  # do not use strided spectrogram

    def get_spectrogram(self) -> Spectrogram | None:
        """Calculate the `Spectrogram` to be plotted.

        This exists as a separate method to allow subclasses to override
        this and not the entire `get_plot` method, e.g. `Coherencegram`.

        This method should not apply the normalisation from `args.norm`.
        """
        args = self.args

        fftlength = float(args.secpfft)
        overlap = fftlength * args.overlap
        self.logger.debug(
            "Calculating spectrogram secpfft: %f, overlap: %f",
            fftlength,
            overlap,
        )

        stride = self.get_stride()

        if stride:
            specgram = self.timeseries[0].spectrogram(
                stride,
                fftlength=fftlength,
                overlap=overlap,
                method=args.method,
                window=args.window,
            )
            nfft = stride * (stride // (fftlength - overlap))
            self.logger.debug(
                "Spectrogram calc, stride: %f, fftlength: %f, "
                "overlap: %f, #fft: %d",
                stride,
                fftlength,
                overlap,
                nfft,
            )
        else:
            specgram = self.timeseries[0].spectrogram2(
                fftlength=fftlength,
                overlap=overlap,
                window=args.window,
            )
            nfft = specgram.shape[0]
            self.logger.debug(
                "HR-Spectrogram calc, fftlength: %f, overlap: %f, #fft: %d",
                fftlength,
                overlap,
                nfft,
            )

        return specgram ** (1/2.)   # ASD

    def make_plot(self) -> Plot | None:
        """Generate the plot from time series and arguments."""
        args = self.args

        # constant input causes unhelpful (weird) error messages
        # translate them to English
        inmin = self.timeseries[0].min().value
        if inmin == self.timeseries[0].max().value:
            if not self.got_error:
                self.logger.error(
                    "Input has constant values [%g]. "
                    "Spectrogram-like products cannot process them.",
                    inmin,
                )
            self.got_error = True
            return None

        # create 'raw' spectrogram
        specgram = self.get_spectrogram()

        # there may be data that can't be processed
        if specgram is None:
            return None

        # apply normalisation
        if args.norm:
            specgram = specgram.ratio(args.norm)

        self.result = specgram

        # -- update plot defaults

        if not args.ymin:
            args.ymin = 1/args.secpfft if args.yscale == "log" else 0

        norm = "log" if args.color_scale == "log" else None
        # vmin/vmax set in scale_axes_from_data()
        return specgram.plot(
            figsize=self.figsize,
            dpi=self.dpi,
            norm=norm,
            cmap=args.cmap,
        )

    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Auto-scale the view based on visible data."""
        args = self.args
        result = cast("Spectrogram", self.result)

        # get tight axes limits from time and frequency Axes
        if args.xmin is None:
            args.xmin = result.xspan[0]
        if args.xmax is None:
            args.xmax = result.xspan[1]
        if args.ymin is None:
            args.ymin = result.yspan[0]
        if args.ymax is None:
            args.ymax = result.yspan[1]

        specgram = result.crop(
            args.xmin,
            min(args.xmax, result.xspan[1]),
        )

        # y axis cannot be cropped if non-linear
        ax.set_ylim(args.ymin, args.ymax)

        # auto scale colours
        if args.norm:
            imin = specgram.value.min()
            imax = specgram.value.max()
        else:
            imin = percentile(specgram.value, 1)
            imax = percentile(specgram.value, 100)
        imin = args.imin if args.imin is not None else imin
        imax = args.imax if args.imax is not None else imax

        if imin == 0 and args.color_scale == "log":
            imin = specgram.value[specgram.value > 0].min()

        self.logger.debug("Colorbar limits set to %f - %f", imin, imax)

        image: AxesImage | Collection
        try:
            image = ax.images[0]
        except IndexError:
            image = ax.collections[0]
        image.set_clim(imin, imax)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.cwb`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from gwpy.table import EventTable

if TYPE_CHECKING:
    from pathlib import Path

    from astropy.table import Table

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


CWB_ASCII = """
# correlation threshold = 0.000000
# network rho threshold = 0.000000
# -/+ - not passed/passed final selection cuts
#  1 - effective correlated amplitude rho
#  2 - correlation coefficient 0/1
#  3 - event norm
#  4 - event DoF
#  5 - correlated amplitude
#  6 - time shift
#  7 - time super shift
#  8 - likelihood
#  9 - normalized chi2
# 10 - cross-power statistic
# 11 - central frequency
# 12 - bandwidth
# 13 - duration
# 14 - number of pixels
# 15 - frequency resolution
# 16 - cwb run number
# 17 - time for L1 detector
# 18 - time for H1 detector
# 19 - sSNR for L1 detector
# 20 - sSNR for H1 detector
# 21 - hrss for L1 detector
# 22 - hrss for H1 detector
# 23 - PHI
# 24 - THETA
# 25 - PSI
+ - 72.5067 0.990 0.965 67.908 72.59   0.000       0 5.3e+03 0.980 152.705  123   67 0.023 1292   0     1 1420000000.2036 1420000000.2059 2.8e+03 2.5e+03 1.9e-22 1.7e-22 300.26 43.43 21.36
+ - 18.1662 0.828 0.723 81.677 19.00   0.000       0 4.1e+02 0.966 26.177  109   66 0.172 1140   0     1 1420000001.1299 1420000001.1361 2.5e+02 1.6e+02 6.2e-23 4.8e-23 274.22 91.79 -74.38
+ - 13.7103 0.940 0.842 31.190 13.73   0.000       0 2.0e+02 0.882 21.139  109   46 0.017 528   0     1 1420000002.7594 1420000002.7623 1.2e+02 8.5e+01 3.7e-23 3.2e-23 266.48 66.68 -79.10
+ - 13.2168 0.908 0.900 30.457 13.49   0.000       0 1.9e+02 1.034 21.953  104   44 0.019 517   0     1 1420000003.2002 1420000003.2079 1.2e+02 7.4e+01 3.7e-23 2.9e-23 276.33 115.94 -78.07
""".strip()  # noqa: E501


@pytest.fixture
def cwb_ascii(tmp_path: Path) -> Path:
    """Create a temporary cWB ASCII file."""
    dat = tmp_path / "EVENTS.txt"
    dat.write_text(CWB_ASCII)
    return dat


@pytest.fixture
def cwb_root(cwb_ascii: Path) -> Path:
    """Create a temporary cWB ROOT file from the ASCII file."""
    tab = EventTable.read(
        cwb_ascii,
        columns=["central frequency", "PHI", "THETA"],
        format="ascii.cwb",
    )
    rootf = cwb_ascii.parent / "cwb.root"
    tab.write(rootf, format="root", treename="waveburst")
    return rootf


def _check_cwb_table(table: Table):
    """Check that a cWB table has been read correctly."""
    assert len(table) == 4
    assert table[0]["central frequency"] == pytest.approx(123)
    assert table[3]["PHI"] == pytest.approx(276.33)


def test_read_cwb_ascii(cwb_ascii: Path):
    """Check that the ``"ascii.cwb"`` file reader works."""
    tab = EventTable.read(cwb_ascii, format="ascii.cwb")
    _check_cwb_table(tab)


def test_read_cwb_ascii_columns_where(cwb_ascii: Path):
    """Check that the ``"ascii.cwb"`` file reader works."""
    columns = ["central frequency", "bandwidth", "duration"]
    tab = EventTable.read(
        cwb_ascii,
        format="ascii.cwb",
        columns=columns,
        where=["PSI > 0"],
    )
    assert set(tab.colnames) == set(columns)
    assert list(tab["bandwidth"]) == [67]  # only one event with PSI > 0


@pytest.mark.requires("uproot")
def test_read_cwb_root(cwb_root: Path):
    """Check that the ``"ascii.root"`` file reader works."""
    tab = EventTable.read(cwb_root, format="root.cwb")
    _check_cwb_table(tab)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.io.scitokens`."""

import os
import time
from contextlib import nullcontext
from unittest import mock

import pytest

from .. import scitokens as io_scitokens

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# Rerun tests when the SciTokens demo token is invalid
rerun_invalid_token = pytest.mark.flaky(
    reruns=2,
    only_rerun=["InvalidTokenError"],
)

# Claims for the demo token
TEST_CLAIMS = {
    "aud": "https://gwpy.example.com",
    "scope": "gwpy.read gwpy.create gwpy.update",
}


@pytest.fixture
def tokenstr():
    """Return a serialized demo token using ``TEST_CLAIMS``."""
    demo = pytest.importorskip("scitokens.utils.demo")
    tok = demo.token(
        TEST_CLAIMS | {"exp": time.time() + 86400},
    )

    # ensure that iat is in the past
    time.sleep(2)

    return tok


@pytest.fixture
def token(tokenstr):
    """Return a demo `scitoken.SciToken` using ``TEST_CLAIMS``."""
    scitokens = pytest.importorskip("scitokens")
    return scitokens.SciToken.deserialize(tokenstr)


@pytest.fixture
def tokenpath(tmp_path):
    """Return a temporary path in which to write a token.

    The token is ***not*** written.
    """
    return tmp_path / "test.token"


@pytest.fixture
def htgettoken_main(tokenstr, tokenpath):
    """Mock `htgettoken.main` to write the demo ``tokenstr`` to ``tokenpath``."""
    # mock the whole htgettoken call
    def htgettoken(*args, **kwargs):  # noqa: ARG001
        with tokenpath.open("w") as file:
            print(tokenstr, file=file)

    with (
        mock.patch("htgettoken.main", side_effect=htgettoken) as mocker,
        mock.patch(
            "gwpy.io.scitokens.tempfile.TemporaryDirectory",
            return_value=nullcontext(tokenpath.parent),
        ),
    ):
        yield mocker


def assert_tokens_equal(result, expected):
    """Assert that token `scitokens.SciToken` objects have the same claims."""
    assert dict(result.claims()) == dict(expected.claims())


# -- find_token

@mock.patch.dict(os.environ)
@rerun_invalid_token
@pytest.mark.requires("igwn_auth_utils")
def test_find_token(token, tokenstr):
    """Test that `find_token` returns the right token."""
    os.environ["BEARER_TOKEN"] = tokenstr
    found = io_scitokens.find_token(
        TEST_CLAIMS["aud"],
        "gwpy.read",
        warn=True,
    )
    assert_tokens_equal(found, token)


@mock.patch.dict(os.environ)
@rerun_invalid_token
@pytest.mark.requires("igwn_auth_utils")
def test_find_token_error(tokenstr):
    """Test error handling in `find_token`."""
    os.environ["BEARER_TOKEN"] = tokenstr
    with (
        pytest.raises(
            RuntimeError,
            match="could not find a valid SciToken",
        ),
        pytest.warns(
            UserWarning,
            match="Validator rejected value of '.*' for claim 'scope'",
        ),
    ):
        io_scitokens.find_token(
            TEST_CLAIMS["aud"],
            "gwpy.delete",
            warn=True,
        )


# -- get_token

@rerun_invalid_token
@pytest.mark.requires("htgettoken", "scitokens", exc_type=(ImportError, OSError))
@pytest.mark.usefixtures("htgettoken_main")
def test_get_token(token, tokenpath):
    """Test `get_scitoken`."""
    # get a new token
    new = io_scitokens.get_scitoken(
        outfile=tokenpath,
        minsecs=600,
        quiet=False,
    )

    # check that we got the right token
    assert_tokens_equal(new, token)


@rerun_invalid_token
@pytest.mark.requires("htgettoken", "scitokens", exc_type=(ImportError, OSError))
def test_get_token_error_systemexit():
    """Test that `get_scitoken` handles `SystemExit` well."""
    with pytest.raises(
        RuntimeError,
        match="htgettoken failed",
    ) as exc_info:
        io_scitokens.get_scitoken(badkwarg=0)
    assert "no such option: --badkwarg" in str(exc_info.getrepr(chain=True))


# -- add_http_authorization_header

@rerun_invalid_token
@pytest.mark.requires("requests_scitokens")
def test_add_http_authorization_header_token(tokenstr, token):
    """Test `add_http_authorization_header`."""
    headers: dict[str, str] = {}
    io_scitokens.add_http_authorization_header(
        headers,
        token=token,
    )
    assert headers["Authorization"] == f"Bearer {tokenstr}"


@rerun_invalid_token
@mock.patch.dict("sys.modules", {"requests_scitokens": None})
def test_add_http_authorization_header_missing_import():
    """Test `add_http_authorization_header` handling of missing import."""
    with pytest.raises(ImportError):
        io_scitokens.add_http_authorization_header({})
    io_scitokens.add_http_authorization_header(headers := {}, error=False)
    assert "Authorization" not in headers


@mock.patch.dict(os.environ)
@pytest.mark.flaky(
    reruns=2,
    # find_token raises a RuntimeError on top of InvalidTokenError
    only_rerun=[RuntimeError],
)
@pytest.mark.requires("igwn_auth_utils", "requests_scitokens")
def test_add_http_authorization_header_find(tokenstr):
    """Test `add_http_authorization_header` calling out to find_token."""
    os.environ["BEARER_TOKEN"] = tokenstr
    headers: dict[str, str] = {}
    io_scitokens.add_http_authorization_header(
        headers,
        audience=TEST_CLAIMS["aud"],
        warn=True,
    )
    assert headers["Authorization"] == f"Bearer {tokenstr}"


@mock.patch.dict(os.environ)
@rerun_invalid_token
@pytest.mark.requires(
    "htgettoken",
    "igwn_auth_utils",
    "requests_scitokens",
    exc_type=(ImportError, OSError),
)
@pytest.mark.usefixtures("htgettoken_main")
def test_add_http_authorization_header_get():
    """Test `add_http_authorization_header` calling out to get_scitoken."""
    os.environ.pop("BEARER_TOKEN", None)
    os.environ.pop("BEARER_TOKEN_FILE", None)
    headers: dict[str, str] = {}
    io_scitokens.add_http_authorization_header(
        headers,
        audience=TEST_CLAIMS["aud"],
    )


@mock.patch("requests_scitokens.HTTPSciTokenAuth.find_token", side_effect=RuntimeError)
@mock.patch("gwpy.io.scitokens.get_scitoken", side_effect=RuntimeError)
@pytest.mark.requires("requests_scitokens")
def test_add_http_authorization_error(*_mocks):
    """Test `add_http_authorization_header` ``error`` keyword behaviour."""
    # check that the right error gets raised
    with pytest.raises(
        RuntimeError,
        match="failed to identify a valid SciToken",
    ):
        io_scitokens.add_http_authorization_header({})

    # check that error=False supresses the errors
    headers: dict[str, str] = {}
    io_scitokens.add_http_authorization_header(headers, error=False)
    assert "Authorization" not in headers



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Accessing and visualising public GW detector data
#################################################

Data from the current generation gravitational wave detectors are published
by |GWOSCl| and freely available to the public.
In this example we demonstrate how to identify times of a published
GW detection event, and to download and visualise detector data.
"""

# %%
# Firstly, we can use the :doc:`gwosc <gwosc:index>` Python package to query
# for the time of the first gravitational-wave detection |GW150914|_:

from gwosc.datasets import event_gps
gps = event_gps("GW150914")

# %%
# GWpy's `TimeSeries` class provides an interface to the public |GWOSC|_
# data in the :meth:`~TimeSeries.get` method; to use it we
# need to first import the `TimeSeries` object:

from gwpy.timeseries import TimeSeries

# %%
# then call the :meth:`~TimeSeries.get` method, passing it the
# prefix for the interferometer we want (`'L1'` here for LIGO-Livingston),
# and the GPS start and stop times of our query (based around the GPS time
# for GW150914):

data = TimeSeries.get("L1", gps-5, gps+5)

# %%
# and then we can make a plot:

plot = data.plot(
    title="LIGO Livingston Observatory data for GW150914",
    ylabel="Strain amplitude",
    color="gwpy:ligo-livingston",
    epoch=gps,
)
plot.show()

# %%
# We can't see anything that looks like a gravitational wave signal in these
# data, the amplitude is dominated by low-frequency detector noise.
# Further filtering is required to be able to identify the GW150914 event
# here, see :ref:`sphx_glr_examples_signal_gw150914.py` for a more in-depth example of
# extracting signals from noise.



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.omega`."""

from contextlib import nullcontext

import h5py
import pytest

from ...testing.utils import assert_table_equal
from ..filter import filter_table
from ..table import EventTable
from .utils import random_table

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.fixture
def snaxtable():
    """Create a table of SNAX-like data."""
    snax = random_table(
        names=[
            "time",
            "snr",
            "frequency",
        ],
        length=100,
    )
    snax["channel"] = "X1:SNAX"
    return snax


@pytest.fixture
def snaxfile(snaxtable, tmp_path):
    """Create a temporary file with example SNAX HDF5 data."""
    tmp = tmp_path / "SNAX-0-0.h5"
    channel = snaxtable[0]["channel"]
    tmptable = snaxtable.copy()
    tmptable.remove_column("channel")
    with h5py.File(tmp, "w") as h5f:
        group = h5f.create_group(channel)
        group.create_dataset(data=tmptable, name="0.0_20.0")
    return tmp


def test_read_snax(snaxtable, snaxfile):
    """Check that we can read a SNAX-format HDF5 file."""
    table = EventTable.read(snaxfile, format="hdf5.snax")
    assert_table_equal(snaxtable, table)


def test_read_snax_channel(snaxtable, snaxfile):
    """Check that we can read a SNAX-format HDF5 file specifying the channel."""
    table = EventTable.read(
        snaxfile,
        format="hdf5.snax",
        channels="X1:SNAX",
    )
    assert_table_equal(snaxtable, table)


def test_read_snax_columns_where(snaxtable, snaxfile):
    """Check that ``where`` and ``columns`` work when reading from a SNAX file."""
    # test with where and columns
    table = EventTable.read(
        snaxfile,
        channels="X1:SNAX",
        format="hdf5.snax",
        where="snr>.5",
        columns=("time", "snr"),
    )
    assert_table_equal(
        table,
        filter_table(snaxtable, "snr>.5")[("time", "snr")],
    )


def test_read_snax_compact(snaxtable, snaxfile):
    """Check that ``columns`` and ``where`` work when reading from a SNAX file."""
    # test compact representation of channel column
    t2 = EventTable.read(snaxfile, compact=True, format="hdf5.snax")

    # group by channel and drop channel column
    tables = {}
    t2 = t2.group_by("channel")
    t2.remove_column("channel")
    for key, group in zip(t2.groups.keys, t2.groups, strict=True):
        channel = t2.meta["channel_map"][key["channel"]]
        tables[channel] = EventTable(group, copy=True)

    # verify table groups are identical
    t_ref = snaxtable.copy().group_by("channel")
    t_ref.remove_column("channel")
    for key, group in zip(t_ref.groups.keys, t_ref.groups, strict=True):
        channel = key["channel"]
        assert_table_equal(group, tables[channel])


_missing_msg = "requested channels not found in SNAX file: 'X1:MISSING'"

@pytest.mark.parametrize(("on_missing", "ctx"), [
    pytest.param(
        "error",
        pytest.raises(ValueError, match=_missing_msg),
        id="error",
    ),
    pytest.param(
        "warn",
        pytest.warns(UserWarning, match=_missing_msg),
        id="warn",
    ),
    pytest.param(
        "ignore",
        nullcontext(),
        id="ignore",
    ),
])
def test_read_snax_errors(on_missing, ctx, snaxtable, snaxfile):
    """Check error handling when reading from a SNAX-format file."""
    missing = ["X1:SNAX", "X1:MISSING"]
    with ctx:
        table = EventTable.read(
            snaxfile,
            channels=missing,
            format="hdf5.snax",
            on_missing=on_missing,
        )
    if on_missing != "error":
        assert_table_equal(snaxtable, table)



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.tex`."""

from __future__ import annotations

from unittest import mock

import pytest

from .. import tex as plot_tex


def _which(arg: str) -> str | None:
    """Fake which to force pdflatex to being not found.

    Parameters
    ----------
    arg : `str`
        The path to find.

    Returns
    -------
    path : `str` or `None`
        If ``arg`` is ``"pdflatex"`` this function returns `None`,
        otherwise ``arg`` is returned verbatim.
    """
    if arg == "pdflatex":
        return None
    return arg


@mock.patch("gwpy.plot.tex.which", _which)
def test_has_tex_missing_exe():
    """Test `has_tex()` handling of missing executables."""
    plot_tex.has_tex.cache_clear()
    assert not plot_tex.has_tex()


@mock.patch("gwpy.plot.tex._test_usetex", side_effect=RuntimeError)
def test_has_tex_bad_latex(mock_test_usetex):
    """Test `has_tex()` handling of a rendering failure."""
    plot_tex.has_tex.cache_clear()
    assert not plot_tex.has_tex()


@mock.patch("gwpy.plot.tex.which", return_value="path")
@mock.patch("gwpy.plot.tex._test_usetex")
def test_has_tex_true(mock_which_, mock_test_usetex):
    """Test `has_tex()` returns `True` when everything works."""
    plot_tex.has_tex.cache_clear()
    assert plot_tex.has_tex()


@pytest.mark.parametrize(("in_", "out"), [
    pytest.param(
        1,
        "1",
        id="1",
    ),
    pytest.param(
        100,
        r"10^{2}",
        id="power of ten",
    ),
    pytest.param(
        -500,
        r"-5\!\!\times\!\!10^{2}",
        id="scientific notation",
    ),
    pytest.param(
        0.00001,
        r"10^{-5}",
        id="negative power of ten",
    ),
])
def test_float_to_latex(in_, out):
    """Test `float_to_latex()`."""
    assert plot_tex.float_to_latex(in_) == out


@pytest.mark.parametrize(("in_", "out"), [
    pytest.param(
        None,
        "",
        id="None",
    ),
    pytest.param(
        "normal text",
        "normal text",
        id="text",
    ),
    pytest.param(
        r"$1 + 2 = 3$",
        r"$1 + 2 = 3$",
        id="mathtex",
    ),
    pytest.param(
        "H1:ABC-DEF_GHI",
        r"H1:ABC-DEF\_GHI",
        id="underscore",
    ),
    pytest.param(
        r"H1:ABC-DEF\_GHI",
        r"H1:ABC-DEF\_GHI",
        id="escaped",
    ),
])
def test_label_to_latex(in_, out):
    """Test `label_to_latex()`."""
    assert plot_tex.label_to_latex(in_) == out
